import torch
import torch.nn as nn
import torch.nn.functional as F

from watch.tasks.rutgers_material_change_detection.models.base_model import BaseFramework


class EarlyFusionMatFramework(BaseFramework):
    def __init__(self, task_mode, encoder, decoder, attention, mat_encoder, mat_embed_dim, mat_integration, **kwargs):
        super(EarlyFusionMatFramework, self).__init__(task_mode, encoder, decoder, attention)

        self.mat_encoder = mat_encoder
        self.mat_integration = mat_integration

        self.mat_feat_proj = nn.Sequential(
            nn.Conv2d(mat_encoder.discretizer.out_feat_dim, mat_embed_dim, kernel_size=1),
            nn.PReLU(),
            nn.Conv2d(mat_embed_dim, mat_embed_dim, kernel_size=1),
        )

    def forward(self, data):
        """Forward pass.

        Args:
            data (dict): [description]

        Returns:
            [type]: [description]
        """

        output = {}

        video = data["video"]  # [batch_size, frames, channels, height, width]

        # Generate material feature masks from frames.
        n_frames = video.shape[1]
        if self.mat_integration == "features":
            comb_frames = []
            for i in range(n_frames):
                mat_feat, _ = self.mat_encoder.encode_mat_feats(video[:, i])

                mat_feat = self.mat_feat_proj(mat_feat)

                # Resize features.
                mat_feat = F.interpolate(mat_feat, size=(video.shape[3], video.shape[4]), mode="nearest")

                # Stack video frames and material features.
                frame = video[:, i]
                comb_frames.append(torch.concat((frame, mat_feat), dim=1))
            video = torch.concat(comb_frames, dim=1)

        elif self.mat_integration == "change_conf":
            mat_confs = []
            for i in range(n_frames):
                _, mat_conf = self.mat_encoder.encode_mat_feats(video[:, i])
                mat_confs.append(mat_conf)

            # mat_change_conf = 1 - torch.max(torch.prod(torch.stack(mat_confs, dim=0), dim=0), dim=1)[0]
            mat_change_conf = 1 - torch.max(torch.mean(torch.stack(mat_confs, dim=0), dim=0), dim=1)[0]
            mat_change_conf = mat_change_conf**4
            # Resize features.
            mat_change_conf = F.interpolate(
                mat_change_conf.unsqueeze(1), size=(video.shape[3], video.shape[4]), mode="bilinear", align_corners=True
            )
            output["mat_change_conf"] = mat_change_conf

            # Stack video frames and material features.
            n_channels = video.shape[2]
            mat_change_frame = mat_change_conf.unsqueeze(1).repeat(1, 1, n_channels, 1, 1)
            video = torch.concat([video, mat_change_frame], dim=1)

            # Stack frames along the channel dimension.
            n_channels = video.shape[1] * video.shape[2]
            video = video.view(video.shape[0], n_channels, video.shape[3], video.shape[4])
        else:
            raise NotImplementedError

        # Pass input into the encoder and decoder.
        feats = self.encoder(video)

        # Compute attention over features.
        if self.attention:
            # TODO: Concatenate relative positional features here.
            feats = self.attention(feats)

        # Pass features through decoder.
        if self.decoder.name == "unet":
            pred = self.decoder(feats, video)
        else:
            pred = self.decoder(feats)

        output[self.task_mode] = pred
        return output



"""
Runs tests on a small region over different known stac collections with
different processing levels to ensure we populate kwcoco metadata correctly.


SeeAlso:
    ../watch/stac/stac_search_builder.py
    ../watch/cli/stac_search.py
    ../watch/demo/demo_region.py
"""

import pytest


def run_metadata_test():
    pytest.skip('unfinished')

    import ubelt as ub
    from watch.demo import demo_region
    from watch.utils import util_gis
    region_fpath = demo_region.demo_smart_region_fpath()
    region_gdf = list(util_gis.coerce_geojson_datas(region_fpath))[0]['data']
    region_row = region_gdf[region_gdf['type'] == 'region'].iloc[0]

    demo_dpath = ub.Path.appdir('watch/demo/datasets/smart_test').ensuredir()
    search_fpath = demo_dpath / 'stac_search.json'

    result_fpath = demo_dpath / (region_row['region_id'] + '.input')

    import watch
    repo_dpath = ub.Path(watch.__file__).parent.parent
    secrets_fpath = repo_dpath / 'secrets/secrets'
    if secrets_fpath.exists():
        # hack for environs
        secret_text = secrets_fpath.read_text()
        for line in secret_text.split('\n'):
            if line.startswith('export SMART_STAC_API_KEY'):
                import os
                key = line.split('=')[1].strip()
                os.environ['SMART_STAC_API_KEY'] = key

    from watch.stac import stac_search_builder
    stac_search_builder.main(cmdline=0, **{
        'start_date': region_row['start_date'],
        'end_date': region_row['end_date'],
        'cloud_cover': 40,
        # 'sensors': 'sentinel-s2-l2a-cogs',
        # 'sensors': 'sentinel-s2-l2a-cogs',
        # 'sensors': 'landsat-c2l2-sr',
        'sensors': 'ta1-ls-acc',
        'out_fpath': search_fpath,
    })

    from watch.cli import stac_search
    stac_search.main(cmdline=0, **{
        'region_file': region_fpath,
        'search_json': search_fpath,
        'mode': 'area',
        'verbose': 100,
        'outfile': result_fpath,
    })

    from watch.cli import prepare_ta2_dataset
    prepare_ta2_dataset.main(cmdline=0, **{
        'dataset_suffix': 'testing',
        's3_fpath': [result_fpath],
        'collated': [False],
        'dvc_dpath': demo_dpath,
        'aws_profile': 'iarpa',
        'region_globstr': region_fpath,
        'site_globstr': None,
        'fields_workers': 0,
        'convert_workers': 0,
        'align_workers': 0,
        'cache': 0,
        'ignore_duplicates': 0,
        'visualize': 0,
        'backend': 'serial',
        'run': 0,
    })



from glob import glob
from PIL import Image
import pickle as pkl
import os
import torch
import numpy as np
import yaml
import torch.nn.functional as F
import torchvision.transforms.functional as FT
import json
import random


def bandwise_norm(image, channel_first=True):

    if channel_first:
        bs, cs, h, w = image.shape
        for b in range(bs):
            for c in range(cs):
                single_channel = image[b, c, :, :]
                max_value = single_channel.max()
                min_value = single_channel.min()
                single_channel_normalized = (single_channel - min_value) / (max_value - min_value)
                image[b, c, :, :] = single_channel_normalized
    else:
        bs, h, w, cs = image.shape
        raise NotImplementedError

    return image


def otsu(image, num=400, get_bcm=False):
    c, h, w = image.shape
    image = image.view(1, -1)
    # print(image.shape)
    max_value = image.max()
    min_value = image.min()
    total_num = h * w
    step_value = (max_value - min_value) / num
    value = min_value + step_value
    best_inter_class_var = 0
    while value <= max_value:
        data_1 = image[image <= value]
        data_2 = image[image > value]
        # print(value)
        if data_1.shape[0] == 0 or data_2.shape[0] == 0:
            value += step_value
            continue
        w1 = data_1.shape[0] / total_num
        w2 = data_2.shape[0] / total_num

        mean_1 = data_1.mean()
        mean_2 = data_2.mean()

        inter_class_var = w1 * w2 * torch.pow((mean_1 - mean_2), 2)
        if best_inter_class_var < inter_class_var:
            best_inter_class_var = inter_class_var
            best_threshold = value
        value += step_value
    print(f"best: {best_threshold} max value: {max_value}")
    return best_threshold


def stad_image(image, channel_first=True, get_params=False, patches=False):
    if channel_first:
        if patches:
            bs, ps, c, h, w = image.shape
            image = image.reshape(bs, ps, c, h * w)  # (bs, ps, c, h*w)
            mean = image.mean(dim=3, keepdims=True)  # (bs, ps, c, 1)
            center = image - mean  # (bs, ps, c, h*w)
            var = torch.var(center, dim=3, keepdims=True)  # (bs, ps, c, h*w])
        else:
            bs, c, h, w = image.shape
            image = image.reshape(bs, c, h * w)  # (bs, c, h*w)
            mean = image.mean(dim=2, keepdims=True)  # (bs, c, 1)
            center = image - mean  # (bs, c, h*w)
            var = torch.var(center, dim=2, keepdims=True)  # (bs, c, h*w])
        # var = torch.sum(torch.pow(center,2), axis=2, keepdims=True) / (h * w) # (bs, c, 1)
        std = torch.sqrt(var)
        nm_image = center / std  # (bs, c, h*w)
        if patches:
            nm_image = nm_image.view(bs, ps, c, h, w)
        else:
            nm_image = nm_image.view(bs, c, h, w)

    if get_params:
        return nm_image, mean, std
    else:
        return nm_image


def denorm(image: torch.Tensor, mean: list = [0.485, 0.456, 0.406], std: list = [0.229, 0.224, 0.225]) -> torch.Tensor:
    """denorm shifted image

    Parameters
    ----------
    image : torch.Tensor
        image to denorm
    mean : list, optional
        mean applied to original image, by default [0.485, 0.456, 0.406]
    std : list, optional
        std applied to original image, by default [0.229, 0.224, 0.225]

    Returns
    -------
    torch.Tensor
        denromed image
    """
    if image.dim() == 3:
        assert image.dim() == 3, "Expected image [CxHxW]"
        assert image.size(0) == 3, "Expected RGB image [3xHxW]"

        for t, m, s in zip(image, mean, std):
            t.mul_(s).add_(m)
    elif image.dim() == 4:
        # batch mode
        assert image.size(1) == 3, "Expected RGB image [3xHxW]"

        for t, m, s in zip((0, 1, 2), mean, std):
            image[:, t, :, :].mul_(s).add_(m)

    return image


def create_dir_if_doesnt_exist(dir_path: str) -> None:
    """creates directory in given path

    Parameters
    ----------
    dir_path : str
        path to be created
    """
    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
    return


def normalize_image(image: np.array) -> np.array:
    """normalize image between 0-1

    Parameters
    ----------
    image : np.array

    Returns
    -------
    np.array
    """
    image_min = np.min(image)
    image_max = np.max(image)
    # print(f"image min: {image_min}, image_max: {image_max}")
    return (image - image_min) / (image_max - image_min + 0.00000000001)


def normalize_image_255(image: np.array) -> np.array:
    """normalizes image between 0-255

    Parameters
    ----------
    image : np.array
        image to be normalized

    Returns
    -------
    np.array
        normalized image
    """
    image_min = np.min(image)
    # image_max = np.max(image)
    # print(f"image min: {image_min}, image_max: {image_max}")
    return (image - image_min) / (255 - image_min + 0.00000000001)


def normalize_dm(image: np.array, confidence_score: float = 0) -> np.array:
    """normalize distance map

    Parameters
    ----------
    image : np.array
        distance map to be nromalized
    confidence_score : float, optional
        confidence score, by default 0

    Returns
    -------
    np.array
        normalized distance map
    """
    image_min = np.min(image)
    image_max = np.max(image)
    # print(f"image min: {image_min}, image_max: {image_max}")
    normalized = (image - image_min) / (image_max - image_min + 0.00000000001)
    normalized += confidence_score
    return normalized


def max_norm(p, version='torch', e=1e-5):
    if version == 'torch':
        if p.dim() == 3:
            C, H, W = p.size()
            p = F.relu(p)
            max_v = torch.max(p.view(C, -1), dim=-1)[0].view(C, 1, 1)
            min_v = torch.min(p.view(C, -1), dim=-1)[0].view(C, 1, 1)
            p = F.relu(p - min_v - e) / (max_v - min_v + e)
        elif p.dim() == 4:
            N, C, H, W = p.size()
            p = F.relu(p)
            max_v = torch.max(p.view(N, C, -1), dim=-1)[0].view(N, C, 1, 1)
            min_v = torch.min(p.view(N, C, -1), dim=-1)[0].view(N, C, 1, 1)
            p = F.relu(p - min_v - e) / (max_v - min_v + e)
    elif version in {'numpy', 'np'}:
        if p.ndim == 3:
            C, H, W = p.shape
            p[p < 0] = 0
            max_v = np.max(p, (1, 2), keepdims=True)
            min_v = np.min(p, (1, 2), keepdims=True)
            p[p < min_v + e] = 0
            p = (p - min_v - e) / (max_v + e)
        elif p.ndim == 4:
            N, C, H, W = p.shape
            p[p < 0] = 0
            max_v = np.max(p, (2, 3), keepdims=True)
            min_v = np.min(p, (2, 3), keepdims=True)
            p[p < min_v + e] = 0
            p = (p - min_v - e) / (max_v + e)
    return p


def batch_crf_inference(img: torch.Tensor, probs: torch.Tensor,
                        t: int = 1, scale_factor: int = 1, labels: int = 21) -> torch.Tensor:
    """crf inference for a batch

    Parameters
    ----------
    img : torch.Tensor
        image
    probs : torch.Tensor
        probablity map
    t : int, optional
    scale_factor : int, optional
    labels : int, optional
        number of labels, by default 21

    Returns
    -------
    torch.Tensor
        batch crf predictions
    """
    bs, c, h, w = probs.shape
    image_npy = img.numpy()
    probs_npy = probs.numpy()
    # preds = torch.zeros((bs, 1, h, w))
    preds_probs = torch.zeros((bs, labels, h, w))
    for b in range(bs):
        b_image = image_npy[b, :, :, :]
        b_probs = probs_npy[b, :, :, :]
        b_image = np.ascontiguousarray(np.transpose(b_image, (1, 2, 0)))
        b_image *= 255
        b_image[b_image > 255] = 255
        b_image[b_image < 0] = 0
        b_image = b_image.astype(np.uint8)
        b_pred = crf_inference(b_image, b_probs, t=t,
                               scale_factor=scale_factor, labels=labels)
        preds_probs[b, :, :, :] = torch.from_numpy(b_pred)
        # print(f"pred min: {b_pred.min()} max: {b_pred.max()}")
        # b_pred = b_pred.max(1)[1]
        # print(b_pred.shape)
        # preds[b,:,:,:] = torch.from_numpy(b_pred)
    # return preds, preds_probs
    return preds_probs


def crf_inference(img: np.array, probs: np.array, t: int = 10,
                  scale_factor: int = 1, labels: int = 21) -> np.array:
    """crf prediction for single image

    Parameters
    ----------
    img : np.array
        image
    probs : np.array
        probability map
    t : int, optional
    scale_factor : int, optional
    labels : int, optional

    Returns
    -------
    np.array
        single crf prediction
    """
    import pydensecrf.densecrf as dcrf
    from pydensecrf.utils import unary_from_softmax

    h, w = img.shape[:2]
    n_labels = labels
    d = dcrf.DenseCRF2D(w, h, n_labels)

    unary = unary_from_softmax(probs)
    unary = np.ascontiguousarray(unary)

    d.setUnaryEnergy(unary)
    d.addPairwiseGaussian(sxy=3 / scale_factor, compat=3)
    d.addPairwiseBilateral(sxy=80 / scale_factor, srgb=13,
                           rgbim=np.copy(img), compat=10)
    Q = d.inference(t)

    return np.array(Q).reshape((n_labels, h, w))


def t2n(x: torch.Tensor) -> np.array:
    """convert tensor to numpy array

    Parameters
    ----------
    x : torch.Tensor
        tensor

    Returns
    -------
    np.array
        numpy array
    """
    x = x.cpu().detach().numpy()
    return x


def mat_to_csv(mat_path: str, save_to: str) -> None:
    """convert mat file to csv

    Parameters
    ----------
    mat_path : str
        mat file path
    save_to : str
        target path for csv file
    """
    import scipy.io
    import pandas as pd
    mat = scipy.io.loadmat(mat_path)
    mat = {k: v for k, v in mat.items() if k[0] != '_'}
    data = pd.DataFrame({k: pd.Series(v[0]) for k, v in mat.items()})
    data.to_csv(save_to)


def load_yaml_as_dict(yaml_path: str) -> dict:
    """load config file as dictionarry

    Parameters
    ----------
    yaml_path : str
        path to yaml file

    Returns
    -------
    dict
    """
    with open(yaml_path, 'r') as f:
        config_dict = yaml.load(f, Loader=yaml.FullLoader)
    return config_dict


def dictionary_contents(path: str, types: list,
                        recursive: bool = False) -> list:
    """extyract dictionary and subdictionary contents

    Parameters
    ----------
    path : str
        path of root
    types : list
        types of files to be extracted
    recursive : bool, optional
        search in subsequent directories, by default False

    Returns
    -------
    list
        list of files with full paths
    """
    files = []
    if recursive:
        path = path + "/**/*"
    for type in types:
        if recursive:
            for x in glob(path + type, recursive=True):
                files.append(os.path.join(path, x))
        else:
            for x in glob(path + type):
                files.append(os.path.join(path, x))
    return files


def save_pickle(object: object, path: str, file_name: str) -> None:
    """save pickle to location

    Parameters
    ----------
    object : object
        object to be saved
    path : str
        path to location
    file_name : str
        name of file to be saved
    """
    full_path = path + file_name + ".pkl"
    with open(full_path, 'wb') as file:
        pkl.dump(object, file)
    return


def load_pickle(path: str) -> object:
    """load pickle

    Parameters
    ----------
    path : str
        path to pickle file

    Returns
    -------
    object
        object loaded
    """
    with open(path, 'rb') as file:
        object = pkl.load(file)
    return object


def load_config_as_dict(path_to_config):
    with open(path_to_config, 'r') as stream:
        try:
            config = yaml.safe_load(stream)
        except yaml.YAMLError as exc:
            print(exc)
    return config


def config_parser(path_to_config, experiment_type):
    if experiment_type.lower() == "training":
        config = yaml.safe_load(open(path_to_config))
        return config

    elif experiment_type.lower() == "testing":
        raise Exception("incomplete parser for testing")


def load_json_as_dict(path_to_json):
    with open(path_to_json) as json_file:
        data = json.load(json_file)
    return data


def random_horizonal_flip(image: Image, mask: Image,
                          points: object = False) -> tuple:
    """random horizontal flip of both image and mask

    Parameters
    ----------
    image : PIL.Image
        image to be flipped
    mask : PIL.Image
        mask to be flipped
    points : bool, optional
        whether points are also provided. Can be either boolean or PIL.Image, by default False

    Returns
    -------
    tuple
        typle of image, mask, and possibly points
    """
    hflip = random.random() < 0.5
    if hflip:
        image = FT.hflip(image)
        mask = FT.hflip(mask)
        if isinstance(points, Image.Image):
            points = FT.hflip(points)
    if isinstance(points, Image.Image):
        return image, mask, points
    return image, mask


def random_vertical_flip(image: Image, mask: Image,
                         points: object = False) -> tuple:
    """random vertical flip of both image and mask

    Parameters
    ----------
    image : PIL.Image
        image to be flipped
    mask : PIL.Image
        mask to be flipped
    points : bool, optional
        whether points are also provided. Can be either boolean or PIL.Image, by default False

    Returns
    -------
    tuple
        typle of image, mask, and possibly points
    """
    vflip = random.random() < 0.5
    if vflip:
        image = FT.vflip(image)
        mask = FT.vflip(mask)
        if isinstance(points, Image.Image):
            points = FT.vflip(points)
    if isinstance(points, Image.Image):
        return image, mask, points
    return image, mask



"""
SeeAlso:
    ../cli/find_dvc.py

TODO:
    - [ ] This needs a better API. Something like the API for git/dvc remotes.

Example Usage:

    # List default directories (hard coded ones that exist)
    python -m watch.cli.find_dvc --command=list

    python -m watch.cli.find_dvc --command=add --name=test --path=$HOME --hardware=hdd

    # List after adding
    python -m watch.cli.find_dvc list

    # Now get one
    python -m watch.cli.find_dvc

    # Force it to recall "test"
    python -m watch.cli.find_dvc --name=test

    # Remove the test dir
    python -m watch.cli.find_dvc --command=remove --name=test

    # Final list
    python -m watch.cli.find_dvc --command=list

    python -m watch.cli.find_dvc --hardware=ssd
    python -m watch.cli.find_dvc --hardware=hdd

Example Usage:

    #### ON PROJECT DATA

    # When you register your drop4 data / experiment paths, the DVC examples in
    # this repo will generally work out of the box. The important part is that
    # your path agrees with the tags used in the examples. Telling the registry
    # if the path lives on an HDD or SSD is also useful.
    geowatch_dvc add my_drop4_data --path=$HOME/Projects/SMART/smart_data_dvc --hardware=hdd --priority=100 --tags=phase2_data
    geowatch_dvc add my_drop4_data --path=$HOME/Projects/SMART/smart_expt_dvc --hardware=hdd --priority=100 --tags=phase2_expt

    # The examples in this repo will generally use this pattern to query for
    # the machine-specific data location. Ensure that these commands work
    # and output the correct paths
    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
    DVC_EXPT_DPATH=$(geowatch_dvc --tags='phase2_expt' --hardware=auto)

    echo "DVC_DATA_DPATH = $DVC_DATA_DPATH"
    echo "DVC_EXPT_DPATH = $DVC_EXPT_DPATH"

"""
import scriptconfig as scfg


class FindDVCConfig(scfg.Config):
    """
    Command line helper to find the path to the watch DVC repo

    Example Usage:
        # List currently known directories
        geowatch_dvc list

        # Add a new path (with optional hardware and tags)
        mkdir -p $HOME/tmp/datadir
        geowatch_dvc add --name=testdir --path=$HOME/tmp/datadir --hardware=hdd --tags=mytag

        # Lookup the newly registered path
        geowatch_dvc find --tags mytag

        # Remove the test entry
        geowatch_dvc remove testdir
    """
    __default__ = {
        'command': scfg.Value('find', help='can be find, set, add, list, or remove', position=1),

        'name': scfg.Value(None, help='specify a name to query or store or remove', position=2),

        'hardware': scfg.Value(None, help='Specify hdd, ssd, etc..., Setable and getable property'),

        'priority': scfg.Value(None, help='Higher is more likely. Setable and getable property'),

        'tags': scfg.Value(None, help='User note. Setable and queryable property'),

        'path': scfg.Value(None, help='The path to the dvc repo. Setable and queryable property'),

        'verbose': scfg.Value(1, help='verbosity mode'),

        'must_exist': scfg.Value('auto', help='if True, filter to only directories that exist. Defaults to false except on "find", which is True.')
    }

    @staticmethod
    def main(cmdline=True, **kwargs):
        from watch.utils import util_data
        from rich import print
        import ubelt as ub

        cli_config = FindDVCConfig(data=kwargs, cmdline=cmdline)
        config = dict(cli_config)

        command = config.pop('command')
        verbose = config.pop('verbose')
        must_exist = config.pop('must_exist')
        if must_exist == 'auto':
            must_exist = command == 'find'

        if verbose > 1:
            print('config = {}'.format(ub.urepr(cli_config, nl=1)))

        registry = util_data.DataRegistry()
        if command == 'list':
            registry.list(**config, must_exist=must_exist)
        elif command == 'add':
            registry.add(**config)
        elif command == 'remove':
            registry.remove(name=config['name'])
        elif command == 'set':
            registry.set(**config)
        elif command == 'find':
            dpath = registry.find(**config, must_exist=must_exist)
            print(dpath)
        else:
            raise KeyError(command)
        return 0


__config__ = FindDVCConfig
_CLI = __config__


if __name__ == '__main__':
    """
    CommandLine:
        smartwatch find_dvc
        python ~/code/watch/watch/cli/find_dvc.py
        python -m watch.cli.find_dvc --register
        python -m watch.cli.find_dvc --mode=list


    """
    __config__.main()



#!/usr/bin/env python3
"""
See Old Script:
    ~/code/watch/scripts/run_stac_to_cropped_kwcoco.py
"""
import os
import subprocess
import json

from watch.cli.baseline_framework_ingress import baseline_framework_ingress, load_input_stac_items  # noqa: 501
from watch.cli.baseline_framework_kwcoco_egress import baseline_framework_kwcoco_egress  # noqa: 501
from watch.cli.ta1_stac_to_kwcoco import ta1_stac_to_kwcoco
from watch.cli import coco_add_watch_fields
from watch.utils.util_framework import download_region
import ubelt as ub
import scriptconfig as scfg


class BASDatasetConfig(scfg.DataConfig):
    """
    Generate cropped KWCOCO dataset for BAS from STAC
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework JSON
            '''))

    input_region_path = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework Region definition JSON
            '''))

    output_path = scfg.Value(None, type=str, position=3, required=True, help='S3 path for output JSON')

    from_collated = scfg.Value(False, isflag=True, help=ub.paragraph(
            '''
            Data to convert has been run through TA-1 collation
            '''))

    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            AWS Profile to use for AWS S3 CLI commands
            '''))

    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='Run AWS CLI commands with --dryrun flag')

    virtual = scfg.Value(False, isflag=True, help=ub.paragraph(
            '''
            Ingress will be virtual (using GDAL's virtual file system)
            '''))

    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
            '''
            S3 Output directory for STAC item / asset egress
            '''))

    requester_pays = scfg.Value(False, isflag=True, short_alias=['r'], help=ub.paragraph(
            '''
            Run AWS CLI commands with `--requestor_payer requester` flag
            '''))

    newline = scfg.Value(False, isflag=True, short_alias=['n'], help=ub.paragraph(
            '''
            Output as simple newline separated STAC items
            '''))

    jobs = scfg.Value(1, type=int, short_alias=['j'], help='Number of jobs to run in parallel')

    dont_recompute = scfg.Value(False, isflag=True, help=ub.paragraph(
            '''
            Will not recompute if output_path already exists
            '''))

    previous_input_path = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            STAC json input file for previous interval
            '''))

    target_gsd = scfg.Value(10, type=int, help='Target GSD of output KWCOCO video space')

    time_combine = scfg.Value(False, help=ub.paragraph(
            '''
            Perform time combine on BAS dataset
            '''))


def main():
    config = BASDatasetConfig.cli(strict=True)
    print('config = {}'.format(ub.urepr(dict(config), nl=1, align=':')))
    run_stac_to_cropped_kwcoco(**config)


def build_combined_kwcoco(input_path,
                          previous_input_path,
                          aws_profile,
                          ta1_cropped_dir,
                          dryrun=False,
                          requester_pays=False,
                          jobs=1,
                          virtual=False,
                          from_collated=False):
    if aws_profile is not None:
        aws_ls_command = ['aws', 's3', '--profile', aws_profile, 'ls']
    else:
        aws_ls_command = ['aws', 's3', 'ls']

    if aws_profile is not None:
        aws_cp_command = ['aws', 's3', '--profile', aws_profile, 'cp']
    else:
        aws_cp_command = ['aws', 's3', 'cp']

    input_stac_items = load_input_stac_items(input_path, aws_cp_command)

    combined_stac_items_path = os.path.join(
        ta1_cropped_dir, 'combined_input_stac_items.jsonl')

    # Confirm that the previous interval input path actually exists on
    # S3 (for first iteration it will not)
    try:
        subprocess.run([*aws_ls_command,
                        previous_input_path], check=True)
    except subprocess.CalledProcessError:
        # If we don't have previous interval input path, set the input
        # as the "combined" for next interval
        with open(combined_stac_items_path, 'w') as f:
            print('\n'.join((json.dumps(item)
                             for item in input_stac_items)), file=f)
        return

    previous_input_stac_items = load_input_stac_items(previous_input_path,
                                                      aws_cp_command)
    input_stac_items.extend(previous_input_stac_items)

    with open(combined_stac_items_path, 'w') as f:
        print('\n'.join((json.dumps(item)
                         for item in input_stac_items)), file=f)

    combined_working_dir = '/tmp/combined'
    os.makedirs(combined_working_dir, exist_ok=True)
    combined_ingress_catalog = baseline_framework_ingress(
        combined_stac_items_path,
        combined_working_dir,
        aws_profile=aws_profile,
        dryrun=dryrun,
        requester_pays=requester_pays,
        relative=False,
        jobs=jobs,
        virtual=virtual)

    # 3. Convert ingressed STAC catalog to KWCOCO
    print("* Converting STAC to KWCOCO *")
    ta1_kwcoco_path_for_sc = os.path.join(combined_working_dir,
                                          'combined_ingress_kwcoco.json')
    ta1_stac_to_kwcoco(combined_ingress_catalog,
                       ta1_kwcoco_path_for_sc,
                       assume_relative=False,
                       populate_watch_fields=True,
                       jobs=jobs,
                       from_collated=from_collated,
                       ignore_duplicates=True)

    return ta1_kwcoco_path_for_sc


def run_stac_to_cropped_kwcoco(input_path,
                               input_region_path,
                               output_path,
                               outbucket,
                               from_collated=False,
                               aws_profile=None,
                               dryrun=False,
                               requester_pays=False,
                               newline=False,
                               jobs=1,
                               virtual=False,
                               dont_recompute=False,
                               previous_input_path=None,
                               target_gsd=10,
                               time_combine=False):
    if aws_profile is not None:
        aws_ls_command = ['aws', 's3', '--profile', aws_profile, 'ls']
    else:
        aws_ls_command = ['aws', 's3', 'ls']

    if dont_recompute:
        try:
            subprocess.run([*aws_ls_command, output_path], check=True)
        except subprocess.CalledProcessError:
            # Continue processing
            pass
        else:
            # If output_path file was there, nothing to do
            return

    # 1. Ingress data
    print("* Running baseline framework ingress *")
    ingress_dir = '/tmp/ingress'
    ingress_catalog = baseline_framework_ingress(
        input_path,
        ingress_dir,
        aws_profile=aws_profile,
        dryrun=dryrun,
        requester_pays=requester_pays,
        relative=False,
        jobs=jobs,
        virtual=virtual)

    # 2. Download and prune region file
    print("* Downloading and pruning region file *")
    local_region_path = '/tmp/region.json'
    local_region_path = download_region(input_region_path,
                                        local_region_path,
                                        aws_profile=aws_profile,
                                        strip_nonregions=True)

    # 3. Convert ingressed STAC catalog to KWCOCO
    print("* Converting STAC to KWCOCO *")
    ta1_kwcoco_path = os.path.join(ingress_dir, 'ingress_kwcoco.json')
    ta1_stac_to_kwcoco(ingress_catalog,
                       ta1_kwcoco_path,
                       assume_relative=False,
                       populate_watch_fields=False,
                       jobs=jobs,
                       from_collated=from_collated,
                       ignore_duplicates=True)
    # Add watch fields
    print("* Adding watch fields *")
    coco_add_watch_fields.main(cmdline=False,
                               src=ta1_kwcoco_path,
                               dst=ta1_kwcoco_path,
                               target_gsd=target_gsd,
                               workers=jobs)

    # `ta1_cropped_dir` is the directory that gets recursively copied
    # up to S3, want to put any kwcoco manifests we may need
    # downstream into this directory.  TODO: rename variable to
    # include something like upload_dir or output_dir
    ta1_cropped_dir = '/tmp/cropped_kwcoco'
    os.makedirs(ta1_cropped_dir, exist_ok=True)

    # 3a. Filter KWCOCO dataset by sensors used for BAS
    print("* Filtering KWCOCO dataset for BAS")
    ta1_bas_kwcoco_path = os.path.join(ta1_cropped_dir,
                                       'kwcoco_for_bas.json')
    subprocess.run(['kwcoco', 'subset',
                    '--src', ta1_kwcoco_path,
                    '--dst', ta1_bas_kwcoco_path,
                    '--absolute', 'False',
                    # '--select_images',
                    # '.sensor_coarse == "L8" or .sensor_coarse == "S2"'
                    ],
                   check=True)

    # 3.1. Combine previous interval `kwcoco_for_sc.json` if provided
    # such that SC has full time range of data to work with
    if previous_input_path is not None:
        combined_kwcoco_path = build_combined_kwcoco(
            input_path,
            previous_input_path,
            aws_profile,
            ta1_cropped_dir,
            dryrun=dryrun,
            requester_pays=requester_pays,
            jobs=jobs,
            virtual=virtual,
            from_collated=from_collated)

        if combined_kwcoco_path is None:
            ta1_kwcoco_path_for_sc = ta1_kwcoco_path
        else:
            ta1_kwcoco_path_for_sc = combined_kwcoco_path
    else:
        ta1_kwcoco_path_for_sc = ta1_kwcoco_path

    # 3a. Filter KWCOCO dataset by sensors used for BAS
    # TODO: move this to run_sc_datagen
    print("* Filtering KWCOCO dataset for SC")
    ta1_sc_kwcoco_path = os.path.join(ta1_cropped_dir,
                                      'kwcoco_for_sc.json')
    subprocess.run(['kwcoco', 'subset',
                    '--src', ta1_kwcoco_path_for_sc,
                    '--dst', ta1_sc_kwcoco_path,
                    '--absolute', 'False',
                    '--select_images',
                    '.sensor_coarse == "WV" or .sensor_coarse == "S2"'],
                   check=True)

    # 4. Crop ingress KWCOCO dataset to region for BAS
    print("* Cropping KWCOCO dataset to region for BAS*")
    ta1_cropped_kwcoco_path = os.path.join(ta1_cropped_dir,
                                           'cropped_kwcoco_for_bas.json')
    # include_channels = 'blue|green|red|nir|swir16|swir22|quality'
    include_channels = 'coastal|blue|green|red|B05|B06|B07|nir|B8A|B09|cirrus|swir16|swir22|pan|quality'
    subprocess.run(['python', '-m', 'watch.cli.coco_align',
                    '--visualize', 'False',
                    '--src', ta1_bas_kwcoco_path,
                    '--dst', ta1_cropped_kwcoco_path,
                    '--regions', local_region_path,
                    '--force_nodata', '-9999',
                    '--include_channels', include_channels,  # noqa
                    '--geo_preprop', 'auto',
                    '--keep', 'none',
                    '--convexify_regions', 'True',
                    '--target_gsd', str(target_gsd),  # noqa
                    '--context_factor', '1',
                    '--force_min_gsd', '10',
                    '--workers', str(jobs),  # noqa: 501
                    '--aux_workers', str(include_channels.count('|') + 1),
                    '--rpc_align_method', 'affine_warp'], check=True)

    # 5. "Clean" dataset
    # Ensure that only data bands are used here (i.e. not quality)
    subprocess.run(['python', '-m', 'watch.cli.coco_clean_geotiffs',
                    '--src', ta1_cropped_kwcoco_path,
                    '--channels', 'coastal|blue|green|red|B05|B06|B07|nir|B8A|B09|cirrus|swir16|swir22',
                    '--prefilter_channels', "red",
                    '--min_region_size', '256',
                    '--nodata_value', '-9999',
                    '--workers', str(jobs),  # noqa: 501
                    ], check=True)

    subprocess.run(['python', '-m', 'watch.cli.coco_clean_geotiffs',
                    '--src', ta1_cropped_kwcoco_path,
                    '--channels', "pan",
                    '--prefilter_channels', "pan",
                    '--min_region_size', '256',
                    '--nodata_value', '-9999',
                    '--workers', str(jobs),  # noqa: 501
                    ], check=True)

    # 6. Do the time_combine for BAS
    from watch.utils.util_yaml import Yaml
    default_time_combine_config = ub.udict(
        time_window='1y',
        channels='coastal|blue|green|red|B05|B06|B07|nir|B8A|B09|cirrus|swir16|swir22|pan',
        resolution='10GSD',
        workers='avail',
        start_time='1970-01-01',
        merge_method='mean',
        assets_dname='raw_bands',
    )
    user_time_combine_config = Yaml.coerce(time_combine)

    if user_time_combine_config not in {False, None}:
        if user_time_combine_config is True:
            user_time_combine_config = {}

        time_combine_config = (default_time_combine_config
                               | user_time_combine_config)

        from watch.cli import coco_time_combine
        preproc_kwcoco_fpath = ub.Path(ta1_cropped_kwcoco_path).augment(
            stemsuffix='_timecombined', ext='.kwcoco.zip', multidot=True)
        coco_time_combine.main(
            cmdline=0,
            input_kwcoco_fpath=ta1_cropped_kwcoco_path,
            output_kwcoco_fpath=preproc_kwcoco_fpath,
            **time_combine_config
        )
        ta1_cropped_kwcoco_path = os.fspath(preproc_kwcoco_fpath)

        # Add watch fields
        print("* Adding watch fields to time combined data *")
        coco_add_watch_fields.main(cmdline=False,
                                   src=ta1_cropped_kwcoco_path,
                                   dst=ta1_cropped_kwcoco_path,
                                   target_gsd=target_gsd,
                                   workers=jobs)

    # 7. Egress (envelop KWCOCO dataset in a STAC item and egress;
    #    will need to recursive copy the kwcoco output directory up to
    #    S3 bucket)
    print("* Egressing KWCOCO dataset and associated STAC item *")
    baseline_framework_kwcoco_egress(ta1_cropped_kwcoco_path,
                                     local_region_path,
                                     output_path,
                                     outbucket,
                                     aws_profile=None,
                                     dryrun=False,
                                     newline=False)


if __name__ == "__main__":
    main()



"""ResNet in PyTorch.
ImageNet-Style ResNet
[1] Kaiming He, Xiangyu Zhang, Shaoqing Ren, Jian Sun
    Deep Residual Learning for Image Recognition. arXiv:1512.03385
Adapted from: https://github.com/bearpaw/pytorch-classification
"""
import torch
import torch.nn as nn
import torch.nn.functional as F


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1, is_last=False):
        super(BasicBlock, self).__init__()
        self.is_last = is_last
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        preact = out
        out = F.relu(out)
        if self.is_last:
            return out, preact
        else:
            return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, in_planes, planes, stride=1, is_last=False):
        super(Bottleneck, self).__init__()
        self.is_last = is_last
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, self.expansion * planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(self.expansion * planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion * planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion * planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        preact = out
        out = F.relu(out)
        if self.is_last:
            return out, preact
        else:
            return out


class ResNet(nn.Module):
    def __init__(self, block, num_blocks, in_channel=3, zero_init_residual=False):
        super(ResNet, self).__init__()
        self.in_planes = 64

        self.conv1 = nn.Conv2d(in_channel, 64, kernel_size=3, stride=1, padding=1,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.layer1 = self._make_layer(block, 64, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 128, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 256, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(block, 512, num_blocks[3], stride=2)
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves
        # like an identity. This improves the model by 0.2~0.3% according to:
        # https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)
                elif isinstance(m, BasicBlock):
                    nn.init.constant_(m.bn2.weight, 0)

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for i in range(num_blocks):
            stride = strides[i]
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)

    def forward(self, x, layer=100):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = self.avgpool(out)
        out = torch.flatten(out, 1)
        return out


def resnet18(**kwargs):
    return ResNet(BasicBlock, [2, 2, 2, 2], **kwargs)


def resnet34(**kwargs):
    return ResNet(BasicBlock, [3, 4, 6, 3], **kwargs)


def resnet50(**kwargs):
    return ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)


def resnet101(**kwargs):
    return ResNet(Bottleneck, [3, 4, 23, 3], **kwargs)


model_dict = {
    'resnet18': [resnet18, 512],
    'resnet34': [resnet34, 512],
    'resnet50': [resnet50, 2048],
    'resnet101': [resnet101, 2048],
}


class LinearBatchNorm(nn.Module):
    """Implements BatchNorm1d by BatchNorm2d, for SyncBN purpose"""

    def __init__(self, dim, affine=True):
        super(LinearBatchNorm, self).__init__()
        self.dim = dim
        self.bn = nn.BatchNorm2d(dim, affine=affine)

    def forward(self, x):
        x = x.view(-1, self.dim, 1, 1)
        x = self.bn(x)
        x = x.view(-1, self.dim)
        return x


class SupConResNet(nn.Module):
    """backbone + projection head"""

    def __init__(self, name='resnet50', head='mlp', feat_dim=128):
        super(SupConResNet, self).__init__()
        model_fun, dim_in = model_dict[name]
        self.encoder = model_fun()
        if head == 'linear':
            self.head = nn.Linear(dim_in, feat_dim)
        elif head == 'mlp':
            self.head = nn.Sequential(
                nn.Linear(dim_in, dim_in),
                nn.ReLU(inplace=True),
                nn.Linear(dim_in, feat_dim)
            )
        else:
            raise NotImplementedError(
                'head not supported: {}'.format(head))

    def forward(self, x):
        feat = self.encoder(x)
        feat = F.normalize(self.head(feat), dim=1)
        return feat


class SupCEResNet(nn.Module):
    """encoder + classifier"""

    def __init__(self, name='resnet50', num_classes=10):
        super(SupCEResNet, self).__init__()
        model_fun, dim_in = model_dict[name]
        self.encoder = model_fun()
        self.fc = nn.Linear(dim_in, num_classes)

    def forward(self, x):
        return self.fc(self.encoder(x))


class LinearClassifier(nn.Module):
    """Linear classifier"""

    def __init__(self, name='resnet50', num_classes=10):
        super(LinearClassifier, self).__init__()
        _, feat_dim = model_dict[name]
        self.fc = nn.Linear(feat_dim, num_classes)

    def forward(self, features):
        return self.fc(features)



import tifffile as tif
import torch
from torch.utils import data
import albumentations as A
import os
import os.path as osp


class S7_sort(data.Dataset):
    normalize_params = [
        [0.16198677, 0.22665408, 0.1745371],
        [0.06108317, 0.06515977, 0.04128775]
    ]

    def __init__(self,
                 crop_size=[128, 128],
                 splits='./spacenet/data/splits_unmasked/',  # unmasked images
                 train=True,
                 normalize=False,
                 num_images=2,
                 display=False
                 ):
        self.display = display
        self.train = train
        self.crop = A.Compose([A.RandomCrop(
            height=crop_size[0], width=crop_size[1])], additional_targets={'image2': 'image'})
        if self.train:
            self.rotate = A.Compose(
                [A.RandomRotate90()], additional_targets={'image2': 'image'})
            self.transforms = A.Compose([
                A.ColorJitter(
                    brightness=0.2,
                    contrast=0.2,
                    saturation=0.2,
                    hue=0.),
            ],
                additional_targets={'image2': 'image'}
            )
        else:
            self.transforms = None

        if train:
            self.images = []
            with open(osp.join(splits, 'spacenet7_train_images.txt'), 'r') as file:
                for row in file:
                    self.images.append(row.strip('\n'))

        else:
            self.images = []
            with open(osp.join(splits, 'spacenet7_val_images.txt'), 'r') as file:
                for row in file:
                    self.images.append(row.strip('\n'))

        self.length = len(self.images)
        self.normalize = normalize
        self.normalization = A.Normalize(
            mean=self.normalize_params[0],
            std=self.normalize_params[1],
            max_pixel_value=1.0)
        self.num_images = num_images

    def __len__(self):
        return (self.length)

    def __getitem__(self, idx):

        im_directory, _ = os.path.split(self.images[idx])

        date1 = (int(self.images[idx][-47:-43]),
                 int(self.images[idx][-42:-40]))

        date2 = date1
        while date2 == date1:
            idx2 = torch.randint(len(os.listdir(im_directory)), (1,))[0]
            im2_path = os.path.join(
                im_directory, sorted(
                    os.listdir(im_directory))[idx2])
            date2 = (int(im2_path[-47:-43]), int(im2_path[-42:-40]))

        image = (tif.imread(self.images[idx])).astype(
            "float32")[:, :, :3] / 255
        image2 = (tif.imread(im2_path)).astype("float32")[:, :, :3] / 255

        crop = self.crop(image=image, image2=image2)

        image = crop['image']
        image2 = crop['image2']

        if self.display:
            image_orig, image2_orig = image, image2

        if self.normalize:
            image = self.normalization(image=image)['image']

        if self.transforms:
            image = self.transforms(image=image)['image']
            image2 = self.transforms(image=image2)['image']

            rotated = self.rotate(image=image, image2=image2)
            image = rotated['image']
            image2 = rotated['image2']

        if self.display:
            return torch.tensor(image).permute(2, 0, 1), torch.tensor(image2).permute(
                2, 0, 1), date1, date2, torch.tensor(image_orig).permute(2, 0, 1), torch.tensor(image2_orig).permute(2, 0, 1)

        item = {
            'image1': torch.tensor(image).permute(2, 0, 1),
            'image2': torch.tensor(image2).permute(2, 0, 1),
            'date1': date1,
            'date2': date2
        }
        return item


class S7_time_aligned(data.Dataset):
    normalize_params = [
        [0.16198677, 0.22665408, 0.1745371],
        [0.06108317, 0.06515977, 0.04128775]
    ]

    def __init__(self,
                 crop_size=[128, 128],
                 splits='./spacenet/data/splits_unmasked/',  # unmasked images
                 train=True,
                 normalize=False
                 ):

        self.train = train
        self.crop = A.Compose([A.RandomCrop(
            height=crop_size[0], width=crop_size[1])], additional_targets={'image2': 'image'})
        if self.train:
            self.rotate = A.Compose(
                [A.RandomRotate90()], additional_targets={'image2': 'image'})
            self.transforms = A.Compose([
                A.ColorJitter(
                    brightness=0.2,
                    contrast=0.2,
                    saturation=0.2,
                    hue=0.),
            ],
                additional_targets={'image2': 'image'}
            )
        else:
            self.transforms = None

        if train:
            self.images = []
            with open(osp.join(splits, 'spacenet7_train_images.txt'), 'r') as file:
                for row in file:
                    self.images.append(row.strip('\n'))

        else:
            self.images = []
            with open(osp.join(splits, 'spacenet7_val_images.txt'), 'r') as file:
                for row in file:
                    self.images.append(row.strip('\n'))

        self.length = len(self.images)
        self.normalize = normalize
        self.normalization = A.Normalize(
            mean=self.normalize_params[0],
            std=self.normalize_params[1],
            max_pixel_value=1.0)

    def __len__(self):
        return self.length

    def __getitem__(self, idx):

        im_directory, _ = os.path.split(self.images[idx])

        idx2 = idx
        while abs(idx2 - idx) < 1:
            idx2 = torch.randint(len(os.listdir(im_directory)), (1,))[0]
        im2_path = os.path.join(
            im_directory, sorted(
                os.listdir(im_directory))[idx2])

        image = (tif.imread(self.images[idx])).astype(
            "float32")[:, :, :3] / 255
        image2 = (tif.imread(im2_path)).astype("float32")[:, :, :3] / 255

        idx_false = idx
        im_false_directory, _ = os.path.split(self.images[idx_false])

        while im_false_directory == im_directory:
            idx_false = torch.randint(self.length, (1,))[0]
            im_false_directory, _ = os.path.split(self.images[idx_false])

        image_false = (tif.imread(self.images[idx2])).astype(
            "float32")[:, :, :3] / 255

        crop = self.crop(image=image, image2=image2)
        crop_false = self.crop(image=image_false)

        image = crop['image']
        image2 = crop['image2']
        image_false = crop_false['image']

        if self.normalize:
            image = self.normalization(image=image)['image']
            image2 = self.normalization(image=image2)['image']
            image_false = self.normalization(image=image_false)['image']

        if self.transforms:
            image = self.transforms(image=image)['image']
            image_false = self.transforms(image=image_false)['image']
            image2 = self.transforms(image=image2)['image']

            rotated = self.rotate(image=image, image2=image2)
            image = rotated['image']
            image2 = rotated['image2']

        date1 = (int(self.images[idx][-47:-43]),
                 int(self.images[idx][-42:-40]))
        date2 = (int(im2_path[-47:-43]), int(im2_path[-42:-40]))

        item = {
            'image1': torch.tensor(image).permute(2, 0, 1),
            'image2': torch.tensor(image2).permute(2, 0, 1),
            'date1': date1,
            'date2': date2
        }
        return item


class S7_change_augmentations(data.Dataset):
    normalize_params = [
        [0.16198677, 0.22665408, 0.1745371],
        [0.06108317, 0.06515977, 0.04128775]]  # fix

    def __init__(
        self,
        splits='./spacenet/data/',
        crop_size=[128, 128],
        normalize=True,
        train=True
    ):

        self.train = train
        self.crop = A.Compose([
            A.RandomCrop(height=crop_size[0], width=crop_size[1])],
            additional_targets={
                'image2': 'image', 'mask2': 'mask'})
        if self.train:
            self.transforms = A.Compose([
                A.RandomRotate90(),
                A.ColorJitter(brightness=0.2, contrast=0.2,
                              saturation=0.2, hue=0.),
            ], additional_targets={
                'image2': 'image', 'mask2': 'mask'})
        else:
            self.transforms = None

        self.normalize = normalize
        self.normalization = A.Normalize(
            mean=self.normalize_params[0],
            std=self.normalize_params[1],
            max_pixel_value=1.0)

        if train:
            self.images = []
            with open(osp.join(splits, 'spacenet7_train_images.txt'), 'r') as file:
                for row in file:
                    self.images.append(row.strip('\n'))
            self.masks = []
            with open(osp.join(splits, 'spacenet7_train_building_masks.txt'), 'r') as file:
                for row in file:
                    self.masks.append(row.strip('\n'))
        else:
            self.images = []
            with open(osp.join(splits, 'spacenet7_val_images.txt'), 'r') as file:
                for row in file:
                    self.images.append(row.strip('\n'))
            self.masks = []
            with open(osp.join(splits, 'spacenet7_val_building_masks.txt'), 'r') as file:
                for row in file:
                    self.masks.append(row.strip('\n'))

        self.num_images = len(self.images)

    def __len__(self):
        return self.num_images

    def __getitem__(self, idx):

        #idx = idx % self.num_images
        im_directory, _ = os.path.split(self.images[idx])
        mask_directory, _ = os.path.split(self.masks[idx])

        idx2 = idx
        while abs(idx2 - idx) < 5:
            idx2 = torch.randint(len(os.listdir(im_directory)), (1,))[0]
        im2_path = os.path.join(
            im_directory, sorted(
                os.listdir(im_directory))[idx2])
        mask2_path = os.path.join(
            mask_directory, sorted(
                os.listdir(mask_directory))[idx2])

        image = (tif.imread(self.images[idx])).astype("float32") / 255
        image2 = (tif.imread(im2_path)).astype("float32") / 255
        # buildings are encoded as 255, bring this down to 1
        buildings = tif.imread(self.masks[idx]).astype("int32") // 255
        buildings2 = tif.imread(mask2_path).astype("int32") // 255

        crop = self.crop(image=image, mask=buildings, image2=image2,
                         mask2=buildings2)

        image = crop['image']
        image2 = crop['image2']
        buildings = crop['mask']
        buildings2 = crop['mask2']

        if self.normalize:
            image = self.normalization(image=image)['image']
            image2 = self.normalization(image=image2)['image']

        if self.transforms:
            transformed = self.transforms(
                image=image[:, :, 0:3], image2=image2[:, :, 0:3],
                mask=buildings, mask2=buildings2)
            image[:, :, 0:3] = transformed["image"]
            buildings = transformed["mask"]
            image2[:, :, 0:3] = transformed["image2"]
            buildings2 = transformed["mask2"]

        date1 = (int(self.masks[idx][-47:-43]), int(self.masks[idx][-42:-40]))
        date2 = (int(mask2_path[-47:-43]), int(mask2_path[-42:-40]))

        assert (mask2_path[-47:-43] + '-' + mask2_path[-42:-40] ==
                im2_path[-47:-43] + '-' + im2_path[-42:-40])

        if date2 < date1:
            date1, date2 = date2, date1
            image, image2 = image2, image
            buildings, buildings2 = buildings2, buildings

        region = self.images[idx][-32:-4]

        item = {
            'image1': torch.tensor(image).permute(2, 0, 1),
            'image2': torch.tensor(image2).permute(2, 0, 1),
            'label1': torch.tensor(buildings),
            'label2': torch.tensor(buildings2),
            'date1': date1,
            'date2': date2,
            'region': region
        }
        return item


class S7_change(data.Dataset):
    normalize_params = [
        [0.16198677, 0.22665408, 0.1745371],
        [0.06108317, 0.06515977, 0.04128775]
    ]  # fix

    def __init__(
        self,
        splits='./spacenet/data/',
        crop_size=[128, 128],
        normalize=True,
        transforms=None,
        train=True
    ):

        self.crop = A.Compose([
            A.RandomCrop(
                height=crop_size[0],
                width=crop_size[1])],
            additional_targets={
                'image2': 'image',
                'mask2': 'mask'})
        if transforms:
            self.transforms = A.Compose(transforms)

        self.normalize = normalize
        self.normalization = A.Normalize(
            mean=self.normalize_params[0],
            std=self.normalize_params[1],
            max_pixel_value=1.0)

        if train:
            self.images = []
            with open(osp.join(splits, 'spacenet7_train_images.txt'), 'r') as file:
                for row in file:
                    self.images.append(row.strip('\n'))
            self.masks = []
            with open(osp.join(splits, 'spacenet7_train_building_masks.txt'), 'r') as file:
                for row in file:
                    self.masks.append(row.strip('\n'))
        else:
            self.images = []
            with open(osp.join(splits, 'spacenet7_val_images.txt'), 'r') as file:
                for row in file:
                    self.images.append(row.strip('\n'))
            self.masks = []
            with open(osp.join(splits, 'spacenet7_val_building_masks.txt'), 'r') as file:
                for row in file:
                    self.masks.append(row.strip('\n'))

        self.num_images = len(self.images)

    def __len__(self):
        return self.num_images

    def __getitem__(self, idx):

        #idx = idx % self.num_images
        im_directory, _ = os.path.split(self.images[idx])
        mask_directory, _ = os.path.split(self.masks[idx])

        idx2 = idx
        while abs(idx2 - idx) < 5:
            idx2 = torch.randint(len(os.listdir(im_directory)), (1,))[0]
        im2_path = os.path.join(
            im_directory, sorted(
                os.listdir(im_directory))[idx2])
        mask2_path = os.path.join(
            mask_directory, sorted(
                os.listdir(mask_directory))[idx2])

        image = (tif.imread(self.images[idx])).astype("float32") / 255
        image2 = (tif.imread(im2_path)).astype("float32") / 255
        # buildings are encoded as 255, bring this down to 1
        buildings = tif.imread(self.masks[idx]).astype("int32") // 255
        buildings2 = tif.imread(mask2_path).astype("int32") // 255

        crop = self.crop(
            image=image,
            mask=buildings,
            image2=image2,
            mask2=buildings2)

        image = crop['image']
        image2 = crop['image2']
        buildings = crop['mask']
        buildings2 = crop['mask2']

        if self.normalize:
            image = self.normalization(image=image)['image']
            image2 = self.normalization(image=image2)['image']

        # if self.transforms:
        #     transformed = self.transforms(image=image, mask=buildings)
        #     image = transformed["image"]
        #     buildings = transformed["mask"]

        date1 = (int(self.masks[idx][-47:-43]), int(self.masks[idx][-42:-40]))
        date2 = (int(mask2_path[-47:-43]), int(mask2_path[-42:-40]))

        assert (mask2_path[-47:-43] + '-' + mask2_path[-42:-40] ==
                im2_path[-47:-43] + '-' + im2_path[-42:-40])

        if date2 < date1:
            date1, date2 = date2, date1
            image, image2 = image2, image
            buildings, buildings2 = buildings2, buildings

        region = self.images[idx][-32:-4]

        item = {
            'image1': torch.tensor(image).permute(2, 0, 1),
            'image2': torch.tensor(image2).permute(2, 0, 1),
            'label1': torch.tensor(buildings),
            'label2': torch.tensor(buildings2),
            'date1': date1,
            'date2': date2,
            'region': region
        }
        return item



"""
THIS IS AN OLD VARIANT OF PIPELINE, but is still used in some places.

For the real mlops pipeline see pipeline_nodes.py
"""
import networkx as nx
import itertools as it
import ubelt as ub


def _iter_paths(paths):
    import os
    # Iterate over output paths sequentially
    if isinstance(paths, (str, os.PathLike)):
        yield paths
    else:
        if isinstance(paths, dict):
            for v in paths.values():
                if isinstance(v, list):
                    yield from v
                else:
                    yield v
        elif isinstance(paths, list):
            yield from paths


def _iter_deps(depends):
    if depends:
        if isinstance(depends, list):
            yield from depends
        else:
            yield depends


class Step:
    """
    Represents a single pipeline step

    Example:
        from watch.mlops.pipeline import Pipeline
        pipeline = Pipeline()
        pipeline.submit(
            name='foo',
            command='read in, write out',
            in_paths=['in'],
            out_paths=['out']
        )


    """
    def __init__(step, name, command, in_paths, out_paths, resources=None,
                 depends=None, enabled=True, stage=None):
        """
        Args:

        """
        if resources is None:
            resources = {}
        step.name = name
        step._command = command

        if isinstance(in_paths, list):
            in_paths = dict(enumerate(in_paths))
        if isinstance(out_paths, list):
            out_paths = dict(enumerate(out_paths))

        step.in_paths = ub.udict(in_paths)
        step.out_paths = ub.udict(out_paths)
        step.resources = ub.udict(resources)
        step.depends = depends
        step.stage = stage
        #
        # Set later
        step.enabled = enabled
        step.will_exist = None
        step.otf_cache = True  # on-the-fly cache checking

    @property
    def node_id(step):
        """
        The experiment manager constructs output paths such that they
        are unique given the specific set of inputs and parameters. Thus
        the output paths are sufficient to determine a unique id per step.
        """
        return step.name + '_' + ub.hash_data(step.out_paths)[0:12]

    @property
    def command(step):
        if step.otf_cache and step.enabled != 'redo':
            return step.test_is_computed_command() + ' || ' + step._command
        else:
            return step._command

    def test_is_computed_command(step):
        test_expr = ' -a '.join(
            [f'-e "{p}"' for p in step.out_paths.values()])
        test_cmd = 'test ' +  test_expr
        return test_cmd

    def _iter_opaths(self):
        return _iter_paths(self.out_paths)

    def _iter_ipaths(self):
        return _iter_paths(self.in_paths)

    @ub.memoize_property
    def does_exist(self):
        # return all(self.out_paths.map_values(lambda p: p.exists()).values())
        return all(ub.Path(p).expand().exists() for p in self._iter_opaths())


class Pipeline:

    def __init__(self):
        self.steps = []

    def submit(self, **kwargs):
        step = Step(**kwargs)
        self.steps.append(step)
        return step

    def _populate_explicit_dependency_queue(self, queue):
        for step in self.steps:
            qdepends = [d._qjob for d in _iter_deps(step.depends)]
            step._qjob = queue.submit(
                command=step.command,
                name=step.name,
                depends=qdepends,
                log=False,
            )

    def _populate_implicit_dependency_queue(self, queue, skip_existing=False):
        g = self.connect_steps(skip_existing=skip_existing)

        common_submitkw = {}

        # queue = cmd_queue.Queue.create(
        #     backend=config['backend'], name='prep-ta2-dataset', size=1, gres=None,
        #     environ=environ)

        # Submit steps to the scheduling queue
        for node in g.graph['order']:
            # Skip duplicate jobs
            step = g.nodes[node]['step']
            if step.node_id in queue.named_jobs:
                continue
            depends = []
            for other, _ in list(g.in_edges(node)):
                dep_step = g.nodes[other]['step']
                if dep_step.enabled:
                    depends.append(dep_step.node_id)

            if step.will_exist and step.enabled:
                step._qjob = queue.submit(
                    command=step.command, name=step.node_id, depends=depends,
                    log=False, **common_submitkw)

    def _update_stage_otf_cache(self, cache):
        """
        We have annotated some steps with stages, build a metagraph based on
        that.

        Args:
            cache (List[str] | bool):
                the list of stages to use the otf_cache on. If True then
                all stages are assumed to be cached. Otherwise any ancestor
                of a cached stage also has its cache enabled.
        """
        g = self._build_step_graph()
        stages = self._build_stage_graph()

        if isinstance(cache, int):
            cache = list(stages.nodes) if cache else []

        # For each requested cached stages, also enable the cache of its
        # ancestors. I.e. for any item where the cache is explicitly enabled,
        # enable all of its dependencies (ancestors) as well.
        for stage_name in cache:
            stages.nodes[stage_name]['cache'] = 1
            for n in nx.ancestors(stages, stage_name):
                stages.nodes[n]['cache'] = 1

        for stage_name, stage_data in stages.nodes(data=True):
            stage_data['label'] = '{stage}:cache={cache}'.format(**stage_data)

        # Update the steps themselves
        for stage_name, stage_data in stages.nodes(data=True):
            for step_name in stage_data['members']:
                g.nodes[step_name]['step'].otf_cache = stage_data['cache']

        if 1:
            from cmd_queue.util.util_networkx import write_network_text
            print('Stage Graph:')
            write_network_text(stages)

    def _build_stage_graph(self):
        g = self._build_step_graph()

        stage_to_steps = ub.ddict(list)
        for step in self.steps:
            stage_to_steps[step.stage].append(step.name)

        # Compute the condensed stage graph
        sccs = list(stage_to_steps.values())
        stages = nx.condensation(g, sccs)
        mapping = {}
        for n, stage_data in stages.nodes(data=True):
            stage = g.nodes[stage_data['members'][0]]['step'].stage
            mapping[n] = stage
            stage_data['stage'] = stage
            stage_data['cache'] = 0
        stages = nx.relabel_nodes(stages, mapping)
        return stages

    def _build_step_graph(self):
        steps = self.steps
        g = nx.DiGraph()
        outputs_to_step = ub.ddict(list)
        inputs_to_step = ub.ddict(list)
        for step in steps:
            for path in step._iter_ipaths():
                inputs_to_step[path].append(step.name)
            for path in step._iter_opaths():
                outputs_to_step[path].append(step.name)
            g.add_node(step.name, step=step)

        inputs_to_step = ub.udict(inputs_to_step)
        outputs_to_step = ub.udict(outputs_to_step)

        common = list((inputs_to_step & outputs_to_step).keys())
        for path in common:
            isteps = inputs_to_step[path]
            osteps = outputs_to_step[path]
            for istep, ostep in it.product(isteps, osteps):
                g.add_edge(ostep, istep)

        #
        # Determine which steps are enabled / disabled
        sorted_nodes = list(nx.topological_sort(g))
        g.graph['order'] = sorted_nodes
        return g

    def connect_steps(self, skip_existing=False):
        """
        Build the graph that represents this pipeline using the inputs /
        outputs to find the dependencies
        """
        # Determine the interaction / dependencies between step inputs /
        # outputs
        g = self._build_step_graph()

        for node in g.graph['order']:
            step = g.nodes[node]['step']
            # if config['skip_existing']:
            ancestors_will_exist = all(
                g.nodes[ancestor]['step'].will_exist
                for ancestor in nx.ancestors(g, step.name)
            )
            if skip_existing and step.enabled != 'redo' and step.does_exist:
                step.enabled = False
            step.will_exist = (
                (step.enabled and ancestors_will_exist) or
                step.does_exist
            )

        if 0:
            from cmd_queue.util.util_networkx import write_network_text
            write_network_text(g)
        return g



r"""
An end-to-end script for calling all the scripts needed to

* Pulls the STAC catalog that points to processed large image tiles
* Creates a virtual Uncropped kwcoco dataset that points to the large image tiles
* Crops the dataset to create an aligned TA2 dataset

See Also:
    ~/code/watch/scripts/prepare_drop3.sh
    ~/code/watch/scripts/prepare_drop4.sh
    ~/code/watch/scripts/prepare_drop5.sh

Example:

    # Create a demo region file, and create vairables that point at relevant
    # paths, which are by default written in your ~/.cache folder
    xdoctest -m watch.demo.demo_region demo_khq_region_fpath
    REGION_FPATH="$HOME/.cache/watch/demo/annotations/KHQ_R001.geojson"
    SITE_GLOBSTR="$HOME/.cache/watch/demo/annotations/KHQ_R001_sites/*.geojson"

    # The "name" of the new dataset
    DATASET_SUFFIX=Demo-TA2-KHQ

    # Set this to where you want to build the dataset
    DEMO_DPATH=$PWD/prep_ta2_demo

    mkdir -p "$DEMO_DPATH"

    # This is a string code indicating what STAC endpoint we will pull from
    SENSORS="sentinel-s2-l2a-cogs"

    # Depending on the STAC endpoint, some parameters may need to change:
    # collated - True for IARPA endpoints, Usually False for public data
    # requester_pays - True for public landsat
    # api_key - A secret for non public data

    export SMART_STAC_API_KEY=""
    export GDAL_DISABLE_READDIR_ON_OPEN=EMPTY_DIR

    # Construct the TA2-ready dataset
    python -m watch.cli.prepare_ta2_dataset \
        --dataset_suffix=$DATASET_SUFFIX \
        --cloud_cover=100 \
        --stac_query_mode=auto \
        --sensors "$SENSORS" \
        --api_key=env:SMART_STAC_API_KEY \
        --collated False \
        --requester_pays=True \
        --dvc_dpath="$DEMO_DPATH" \
        --aws_profile=iarpa \
        --region_globstr="$REGION_FPATH" \
        --site_globstr="$SITE_GLOBSTR" \
        --fields_workers=8 \
        --convert_workers=8 \
        --align_workers=26 \
        --cache=0 \
        --ignore_duplicates=1 \
        --target_gsd=30 \
        --visualize=True \
        --max_products_per_region=10 \
        --serial=True --run=0

    smartwatch visualize $HOME/data/dvc-repos/smart_watch_dvc/Aligned-Drop2-TA1-2022-02-24/data.kwcoco_c9ea8bb9.json

TODO:
    handl GE01 and WV01 platforms

"""


import scriptconfig as scfg
import ubelt as ub
import warnings
from cmd_queue.cli_boilerplate import CMDQueueConfig


class PrepareTA2Config(CMDQueueConfig):
    __default__ = {
        'dataset_suffix': scfg.Value(None, help=''),

        'cloud_cover': scfg.Value(10, help='maximum cloud cover percentage (ignored if s3_fpath given)'),
        'sensors': scfg.Value("L2", help='(ignored if s3_fpath given)'),
        'max_products_per_region': scfg.Value(None, help='does uniform affinity sampling over time to filter down to this many results per region'),

        'stac_query_mode': scfg.Value(None, help='if set to auto we try to make the .input files. Mutex with s3_fpath', group='stac'),
        'api_key': scfg.Value('env:SMART_STAC_API_KEY', help='The API key or where to get it (ignored if s3_fpath given)', group='stac'),

        'separate_region_queues': scfg.Value(True, help='if True, create jobs for each region separately. This option to disable this may be removed in the future.'),
        'separate_align_jobs': scfg.Value(True, help='if True, perform alignment for each region in its own job. The option to disable this may be removed in the future.'),

        's3_fpath': scfg.Value(None, nargs='+', help='A list of .input files which were the results of an existing stac query. Mutex with stac_query_* args. Mutex with sensors.', group='stac'),
        'aws_profile': scfg.Value('iarpa', help='AWS profile to use for remote data access', group='stac'),

        'out_dpath': scfg.Value('auto', help='This is the path that all resulting files will be written to. Defaults the the phase2 DATA_DVC_DPATH', alias=['dvc_dpath']),

        'collated': scfg.Value([True], nargs='+', help='set to false if the input data is not collated'),
        'queue_name': scfg.Value('prep-ta2-dataset', help='name for the command queue', group='cmd_queue'),

        'max_regions': None,

        'query_workers': scfg.Value('0', help='workers for STAC search'),
        'convert_workers': scfg.Value('min(avail,8)', help='workers for stac-to-kwcoco script'),
        'fields_workers': scfg.Value('min(avail,max(all/2,8))', help='workers for add-watch-fields script'),

        'align_workers': scfg.Value(0, help='primary workers for align script', group='align'),
        'align_aux_workers': scfg.Value(0, help='threads per align process (typically set this to 0)', group='align'),

        'ignore_duplicates': scfg.Value(1, help='workers for align script'),

        'visualize': scfg.Value(0, isflag=1, help='if True runs visualize'),
        'visualize_only_boxes': scfg.Value(True, isflag=1, help='if False will draw full polygons'),

        'verbose': scfg.Value(0, help='help control verbosity (just align for now)'),

        # '--requester_pays'
        'requester_pays': scfg.Value(0, help='if True, turn on requester_pays in ingress. Needed for official L1/L2 catalogs.'),

        'debug': scfg.Value(False, isflag=1, help='if enabled, turns on debug visualizations'),
        'select_images': scfg.Value(False, help='if enabled only uses select images'),


        'include_channels': scfg.Value(None, help='specific channels to use in align crop', group='align'),
        'exclude_channels': scfg.Value(None, help='specific channels to NOT use in align crop', group='align'),
        'target_gsd': scfg.Value(10, group='align'),
        'force_min_gsd': scfg.Value(None, group='align'),
        'align_keep': scfg.Value('img', choices=['img', 'img-roi', 'none', None], help='if the coco align script caches or recomputes images / rois', group='align'),
        'force_nodata': scfg.Value(None, help='if specified, forces nodata to this value', group='align'),

        'splits': scfg.Value(False, isflag=1, help='if True do splits'),

        'region_globstr': scfg.Value('annotations/region_models', help='region model globstr (relative to the dvc path, unless absolute or prefixed by "./")'),
        'site_globstr': scfg.Value('annotations/site_models', help='site model globstr (relative to the dvc path, unless absolute or prefixed by "./")'),

        'propogate_strategy': scfg.Value('NEW-SMART', help='changes propogation behavior'),

        'remove_broken': scfg.Value(True, isflag=1, help='if True, will remove any image that fails population (e.g. caused by a 404)'),

        'cache': scfg.Value(1, isflag=1, help='If 1 or 0 globally enable/disable caching. If a comma separated list of strings, only cache those stages', group='queue-related'),
        'skip_existing': scfg.Value(False, help='Unlike cache=1, which checks for file existence at runtime, this will explicitly not submit any job with a product that already exist', group='queue-related'),

        'rpc_align_method': scfg.Value('orthorectify', help=ub.paragraph(
            '''
            Can be one of:
                (1) orthorectify - which uses gdalwarp with -rpc if available
                    otherwise falls back to affine transform,
                (2) affine_warp - which ignores RPCs and uses the affine
                    transform in the geotiff metadata.
            '''
        )),

        'hack_lazy': scfg.Value(False, isflag=True, help=ub.paragraph(
            '''
            Hack lazy is a proof of concept with the intent on speeding up the
            download / cropping of data by flattening the gdal processing into
            a single queue of parallel processes executed via a command queue.

            By running once with this flag on, it will execute the command
            queue, and then running again, it should see all of the data as
            existing and construct the aligned kwcoco dataset as normal.
            ''')),
    }


__config__ = PrepareTA2Config


def main(cmdline=False, **kwargs):
    """
    Example:
        from watch.cli.prepare_ta2_dataset import *  # NOQA
        cmdline = False
        kwargs = {
            'dataset_suffix': 'TA1_FULL_SEQ_KR_S001_CLOUD_LT_10',
            's3_fpath': 's3://kitware-smart-watch-data/processed/ta1/eval2/master_collation_working/KR_S001.unique.fixed_ls_ids.cloudcover_lt_10.output',
        }

        kwargs = {
            'dataset_suffix': 'Drop2-2022-02-04',
            's3_fpath': 's3://kitware-smart-watch-data/processed/ta1/drop2_20220204/PE/coreg_and_brdf/watch-coreg-and-brdf_PE.output',
        }

    """
    config = PrepareTA2Config.cli(cmdline=cmdline, data=kwargs, strict=True)
    from watch.utils import slugify_ext
    from watch.utils import util_gis
    import rich
    rich.print('config = {}'.format(ub.urepr(dict(config), nl=1)))

    out_dpath = config['out_dpath']
    if out_dpath == 'auto':
        import watch
        out_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
    out_dpath = ub.Path(out_dpath)

    aws_profile = config['aws_profile']

    aligned_bundle_name = f'Aligned-{config["dataset_suffix"]}'
    uncropped_bundle_name = f'Uncropped-{config["dataset_suffix"]}'

    uncropped_dpath = out_dpath / uncropped_bundle_name
    uncropped_query_dpath = uncropped_dpath / '_query/items'

    uncropped_ingress_dpath = uncropped_dpath / 'ingress'

    aligned_kwcoco_bundle = out_dpath / aligned_bundle_name

    uncropped_dpath = uncropped_dpath.shrinkuser(home='$HOME')
    uncropped_query_dpath = uncropped_query_dpath.shrinkuser(home='$HOME')
    uncropped_query_dpath = uncropped_query_dpath.shrinkuser(home='$HOME')
    uncropped_query_dpath = uncropped_query_dpath.shrinkuser(home='$HOME')
    uncropped_ingress_dpath = uncropped_ingress_dpath.shrinkuser(home='$HOME')
    aligned_kwcoco_bundle = aligned_kwcoco_bundle.shrinkuser(home='$HOME')

    # Ignore these regions (only works in separate region queue mode)
    region_id_blocklist = {
        # 'IN_C000',  # bad files
        # 'ET_C000',  # 404 errors
    }

    # Job environments are specific to single jobs
    job_environs = [
        # 'PROJ_DEBUG=3',
        f'AWS_DEFAULT_PROFILE={aws_profile}',
    ]
    if config['requester_pays']:
        job_environs.append("AWS_REQUEST_PAYER='requester'")
    job_environ_str = ' '.join(job_environs)
    if job_environ_str:
        job_environ_str += ' '

    def _coerce_globstr(p):
        if not p:
            return None
        globstr = ub.Path(p)
        if str(globstr).startswith('./'):
            final_globstr = globstr
        else:
            final_globstr = out_dpath / globstr
        final_globstr = final_globstr.shrinkuser(home='$HOME')
        return final_globstr

    # region_models = list(region_dpath.glob('*.geojson'))
    final_region_globstr = _coerce_globstr(config['region_globstr'])
    final_site_globstr = _coerce_globstr(config['site_globstr'])

    # Global environs are given to all jobs
    api_key = config['api_key']
    environ = {}
    # https://trac.osgeo.org/gdal/wiki/ConfigOptions#GDAL_DISABLE_READDIR_ON_OPEN
    environ['GDAL_DISABLE_READDIR_ON_OPEN'] = 'EMPTY_DIR'
    if api_key is not None and api_key.startswith('env:'):
        import os
        # NOTE!!!
        # THIS WILL LOG YOUR SECRET KEY IN PLAINTEXT!!!
        # TODO: figure out how to pass the in-environment secret key
        # to the tmux sessions.
        api_key_name = api_key[4:]
        api_key_val = os.environ.get(api_key_name, None)
        if api_key_val is None:
            warnings.warn('The requested API key was not available')
        else:
            environ[api_key_name] = api_key_val

    default_collated = config['collated'][0]

    # The pipeline is a layer on top of cmd-queue that will handle connecting
    # inputs / outputs for us. At the time of writing we are still doing that
    # explicitly, but we could remove that code and just rely on implicit
    # dependencies based in in / out paths.
    from watch.mlops.old import pipeline_v1
    pipeline = pipeline_v1.Pipeline()

    stac_jobs = []
    # base_mkdir_job = pipeline.submit(f'mkdir -p "{uncropped_query_dpath}"', name='mkdir-base')
    if config['stac_query_mode'] == 'auto':
        # Each region gets their own job in the queue
        if config['separate_region_queues']:

            # Note: this requires the annotation files to exist on disk.  or we
            # have to write a mechanism that lets the explicit relative path be
            # specified.
            region_file_fpaths = util_gis.coerce_geojson_paths(final_region_globstr.expand())

            if final_site_globstr:
                region_site_fpaths = util_gis.coerce_geojson_paths(final_site_globstr.expand())
            else:
                region_site_fpaths = []

            # Assign site models to region files
            ASSIGN_BY_FPATH = True
            if ASSIGN_BY_FPATH:
                # This is not robust, but it doesn't require touching the disk
                region_id_to_fpath = {p.stem: p for p in region_file_fpaths}
                site_id_to_fpath = {p.stem: p for p in region_site_fpaths}
                region_id_to_site_fpaths = ub.ddict(list)
                for site_id, site_fpaths in site_id_to_fpath.items():
                    region_id, site_num = site_id.rsplit('_', maxsplit=1)
                    region_id_to_site_fpaths[region_id].append(site_fpaths)

                if 1:
                    regions_without_sites = set(region_id_to_fpath) - set(region_id_to_site_fpaths)
                    sites_without_regions = set(region_id_to_site_fpaths) - set(region_id_to_fpath)
                    regions_without_sites_str = slugify_ext.smart_truncate(ub.urepr(regions_without_sites, nl=1), max_length=1000, head="~~\n~~\n", tail="\n~~\n~~")
                    sites_without_regions_str = slugify_ext.smart_truncate(ub.urepr(sites_without_regions, nl=1), max_length=1000, head="~~\n~~\n", tail="\n~~\n~~")
                    print(f'regions_without_sites={regions_without_sites_str}')
                    print(f'sites_without_regions={sites_without_regions_str}')
            else:
                raise NotImplementedError(
                    'TODO: implement more robust alternative that reads '
                    'file data to make assignment if needed')

            if config['max_regions'] is not None:
                region_file_fpaths = region_file_fpaths[:config['max_regions']]

            print(f'region_file_fpaths={slugify_ext.smart_truncate(ub.urepr(sorted(region_file_fpaths), nl=1), max_length=1000)}')
            for region_id, region_fpath in region_id_to_fpath.items():
                if region_id in region_id_blocklist:
                    continue

                region_inputs_fpath = (uncropped_query_dpath / (region_id + '.input')).shrinkuser(home='$HOME')
                final_region_fpath = region_fpath.shrinkuser(home='$HOME')

                stac_search_job = pipeline.submit(
                    command=ub.codeblock(
                        rf'''
                        python -m watch.cli.stac_search \
                            --region_file "{final_region_fpath}" \
                            --search_json "auto" \
                            --cloud_cover "{config['cloud_cover']}" \
                            --sensors "{config['sensors']}" \
                            --api_key "{config['api_key']}" \
                            --query_workers "{config['query_workers']}" \
                            --max_products_per_region "{config['max_products_per_region']}" \
                            --append_mode=False \
                            --mode area \
                            --verbose 2 \
                            --outfile "{region_inputs_fpath}"
                        '''),
                    name=f'stac-search-{region_id}',
                    depends=[],
                    in_paths={
                        'final_region_fpath': final_region_fpath,
                    },
                    out_paths={
                        'region_inputs_fpath': region_inputs_fpath,
                    },
                    stage='stac_search',
                )
                # cache_prefix = f'[[ -f {region_inputs_fpath} ]] || ' if stages.nodes['stac_search']['cache'] else ''

                stac_jobs.append({
                    'name': region_id,
                    'job': stac_search_job,
                    'inputs_fpath': region_inputs_fpath,
                    'region_globstr': final_region_fpath,
                    'collated': default_collated,
                })
        else:
            warnings.warn(
                'It is usually faster to split the queue amongst regions')
            # All region queries are executed simultaniously and put into a
            # single inputs file. The advantage here is we dont need to know
            # how many regions there are beforehand.
            combined_inputs_fpath = (uncropped_query_dpath / (f'combo_query_{config["dataset_suffix"]}.input')).shrinkuser(home='$HOME')

            combo_stac_search_job = pipeline.submit(
                command=ub.codeblock(
                    rf'''
                    python -m watch.cli.stac_search \
                    --region_globstr "{final_region_globstr}" \
                    --search_json "auto" \
                    --cloud_cover "{config['cloud_cover']}" \
                    --sensors "{config['sensors']}" \
                    --api_key "{config['api_key']}" \
                    --max_products_per_region "{config['max_products_per_region']}" \
                    --append_mode=False \
                    --mode area \
                    --verbose 2 \
                    --outfile "{combined_inputs_fpath}"
                    '''),
                name='stac-search', depends=[],
                in_paths={
                    'final_region_globstr': final_region_globstr,
                },
                out_paths={
                    'combined_inputs_fpath': combined_inputs_fpath,
                },
                stage='stac_search',
            )

            stac_jobs.append({
                'name': 'combined',
                'job': combo_stac_search_job,
                'region_globstr': final_region_globstr,
                'inputs_fpath': combined_inputs_fpath,
                'collated': default_collated,
            })
    else:
        s3_fpath_list = config['s3_fpath']
        collated_list = config['collated']
        if len(collated_list) != len(s3_fpath_list):
            print('Indicate if each s3 path is collated or not')
        stac_jobs = []
        for s3_fpath, collated in zip(s3_fpath_list, collated_list):
            stac_jobs.append({
                'job': None,
                'name': ub.Path(s3_fpath).stem,
                'inputs_fpath': s3_fpath,
                'region_globstr': final_region_globstr,
                'collated': collated,
            })

    uncropped_fielded_jobs = []
    for stac_job in stac_jobs:
        s3_fpath = ub.Path(stac_job['inputs_fpath'])
        s3_name = stac_job['name']
        parent_job = stac_job['job']
        collated = stac_job['collated']
        uncropped_query_fpath = uncropped_query_dpath / s3_fpath.name
        uncropped_query_fpath = uncropped_query_fpath.shrinkuser(home='$HOME')

        uncropped_catalog_fpath = uncropped_ingress_dpath / f'catalog_{s3_name}.json'
        uncropped_ingress_dpath = uncropped_ingress_dpath.shrinkuser(home='$HOME')

        if not str(s3_fpath).startswith('s3'):
            # Don't really need to copy anything in this case.
            uncropped_query_fpath = s3_fpath
            grab_job = parent_job
            # grab_job = pipeline.submit(ub.codeblock(
            #     f'''
            #     # GRAB Input STAC List
            #     {cache_prefix}cp "{s3_fpath}" "{uncropped_query_dpath}"
            #     '''), depends=parent_job, name=f'psudo-s3-pull-inputs-{s3_name}')
        else:
            grab_job = pipeline.submit(command=ub.codeblock(
                f'''
                # GRAB Input STAC List
                aws s3 --profile {aws_profile} cp "{s3_fpath}" "{uncropped_query_dpath}"
                '''), depends=parent_job, name=f's3-pull-inputs-{s3_name}',
                in_paths={
                    's3_fpath': s3_fpath,
                },
                out_paths={
                    'uncropped_query_fpath': uncropped_query_fpath,
                },
                stage='grab',
            )

        ingress_options = [
            '--virtual',
        ]
        if config['requester_pays']:
            ingress_options.append('--requester_pays')
        ingress_options_str = ' '.join(ingress_options)

        ingress_job = pipeline.submit(command=ub.codeblock(
            rf'''
            python -m watch.cli.baseline_framework_ingress \
                --aws_profile {aws_profile} \
                --jobs avail \
                {ingress_options_str} \
                --outdir "{uncropped_ingress_dpath}" \
                --catalog_fpath "{uncropped_catalog_fpath}" \
                "{uncropped_query_fpath}"
            '''), depends=[grab_job], name=f'baseline_ingress-{s3_name}',
            in_paths={
                'uncropped_query_fpath': uncropped_query_fpath,
            },
            out_paths={
                'uncropped_catalog_fpath': uncropped_catalog_fpath,
            },
            stage='catalog',
        )

        uncropped_kwcoco_fpath = uncropped_dpath / f'data_{s3_name}.kwcoco.json'
        uncropped_kwcoco_fpath = uncropped_kwcoco_fpath.shrinkuser(home='$HOME')

        convert_options = []
        if collated:
            convert_options.append('--from-collated')
        if config['ignore_duplicates']:
            convert_options.append('--ignore_duplicates')
        convert_options_str = ' '.join(convert_options)

        convert_job = pipeline.submit(
            command=ub.codeblock(
                rf'''
                {job_environ_str}python -m watch.cli.ta1_stac_to_kwcoco \
                    "{uncropped_catalog_fpath}" \
                    --outpath="{uncropped_kwcoco_fpath}" \
                    {convert_options_str} \
                    --jobs "{config['convert_workers']}"
                '''),
            depends=[ingress_job],
            name=f'ta1_stac_to_kwcoco-{s3_name}',
            in_paths={
                'uncropped_catalog_fpath': uncropped_catalog_fpath,
            },
            out_paths={
                'uncropped_kwcoco_fpath': uncropped_kwcoco_fpath,
            },
            stage='uncropped_kwcoco',
        )

        uncropped_fielded_kwcoco_fpath = uncropped_dpath / f'data_{s3_name}_fielded.kwcoco.json'
        uncropped_fielded_kwcoco_fpath = uncropped_fielded_kwcoco_fpath.shrinkuser(home='$HOME')

        add_fields_job = pipeline.submit(command=ub.codeblock(
            rf'''
            {job_environ_str}python -m watch.cli.coco_add_watch_fields \
                --src "{uncropped_kwcoco_fpath}" \
                --dst "{uncropped_fielded_kwcoco_fpath}" \
                --enable_video_stats=False \
                --overwrite=warp \
                --target_gsd={config['target_gsd']} \
                --remove_broken={config['remove_broken']} \
                --workers="{config['fields_workers']}"
            '''),
            depends=convert_job, name=f'coco_add_watch_fields-{s3_name}',
            in_paths={
                'uncropped_kwcoco_fpath': uncropped_kwcoco_fpath,
            },
            out_paths={
                'uncropped_fielded_kwcoco_fpath': uncropped_fielded_kwcoco_fpath,
            },
            stage='uncropped_feilds',
        )

        uncropped_fielded_jobs.append({
            'name': stac_job['name'],
            'job': add_fields_job,
            'uncropped_fielded_fpath': uncropped_fielded_kwcoco_fpath,
            'region_globstr': stac_job['region_globstr'],
        })

    if config['separate_align_jobs']:
        alignment_input_jobs = []
        # TODO: make use of region_id_to_site_fpaths
        # to build a better site globstr (might need to write a tempoaray file
        # to point to)
        final_site_globstr = _coerce_globstr(config['site_globstr'])
        for info in uncropped_fielded_jobs:
            toalign_info = info.copy()
            name = toalign_info['name'] = info['name']
            toalign_info['aligned_imgonly_fpath'] = aligned_kwcoco_bundle / f'imgonly-{name}.kwcoco.json'
            toalign_info['aligned_imganns_fpath'] = aligned_kwcoco_bundle / f'imganns-{name}.kwcoco.json'
            # TODO: take only the corresponding set of site models here.
            toalign_info['site_globstr'] = final_site_globstr
            toalign_info['region_globstr'] = info['region_globstr']
            alignment_input_jobs.append(toalign_info)
    else:
        # Do a noop for this case? Or make it fast in kwcoco itself?
        uncropped_coco_paths = [d['uncropped_fielded_fpath'] for d in uncropped_fielded_jobs]
        union_depends_jobs = [d['job'] for d in uncropped_fielded_jobs]
        union_suffix = ub.hash_data([p.name for p in uncropped_coco_paths])[0:8]
        uncropped_final_kwcoco_fpath = uncropped_dpath / f'data_{union_suffix}.kwcoco.json'
        uncropped_final_kwcoco_fpath = uncropped_final_kwcoco_fpath.shrinkuser(home='$HOME')
        uncropped_multi_src_part = ' '.join(['"{}"'.format(p) for p in uncropped_coco_paths])
        union_job = pipeline.submit(command=ub.codeblock(
            rf'''
            # COMBINE Uncropped datasets
            {job_environ_str}python -m kwcoco union \
                --src {uncropped_multi_src_part} \
                --dst "{uncropped_final_kwcoco_fpath}"
            '''), depends=union_depends_jobs, name='kwcoco-union',
            in_paths={
                'uncropped_coco_paths': uncropped_coco_paths,
            },
            out_paths={
                'uncropped_final_kwcoco_fpath': uncropped_final_kwcoco_fpath,
            },
            stage='union_uncropped_feilds',
        )
        uncropped_final_jobs = [union_job]

        final_site_globstr = _coerce_globstr(config['site_globstr'])
        alignment_input_jobs = [{
            'name': f'align-{union_suffix}',
            'uncropped_fielded_fpath': uncropped_final_kwcoco_fpath,
            'aligned_imgonly_fpath': aligned_kwcoco_bundle / 'imgonly.kwcoco.json',
            'aligned_imganns_fpath': aligned_kwcoco_bundle / 'imganns.kwcoco.json',
            'region_globstr': final_region_globstr,
            'site_globstr': final_site_globstr,
            'job': uncropped_final_jobs,
        }]

    alignment_jobs = []
    for info in alignment_input_jobs:
        name = info['name']
        uncropped_fielded_fpath = info['uncropped_fielded_fpath']
        aligned_imgonly_fpath = info['aligned_imgonly_fpath']
        aligned_imganns_fpath = info['aligned_imganns_fpath']
        region_globstr = info['region_globstr']
        site_globstr = info['site_globstr']
        parent_job = info['job']

        # cache_crops = 1
        # if cache_crops:
        #     align_keep = 'img'
        #     # align_keep = 'roi-img'
        # else:
        #     align_keep = 'none'

        debug_valid_regions = config.debug
        align_visualize = config.debug
        include_channels = config.include_channels
        exclude_channels = config.exclude_channels

        align_job = pipeline.submit(command=ub.codeblock(
            rf'''
            # MAIN WORKHORSE CROP IMAGES
            # Crop big images to the geojson regions
            {job_environ_str}python -m watch.cli.coco_align \
                --src "{uncropped_fielded_fpath}" \
                --dst "{aligned_imgonly_fpath}" \
                --regions "{region_globstr}" \
                --context_factor=1 \
                --geo_preprop=auto \
                --keep={config.align_keep} \
                --force_nodata={config.force_nodata} \
                --include_channels="{include_channels}" \
                --exclude_channels="{exclude_channels}" \
                --visualize={align_visualize} \
                --debug_valid_regions={debug_valid_regions} \
                --rpc_align_method {config.rpc_align_method} \
                --verbose={config.verbose} \
                --aux_workers={config.align_aux_workers} \
                --target_gsd={config.target_gsd} \
                --force_min_gsd={config.force_min_gsd} \
                --workers={config.align_workers} \
                --hack_lazy={config.hack_lazy}
            '''),
            depends=parent_job,
            name=f'align-geotiffs-{name}',
            in_paths={
                'uncropped_fielded_fpath': uncropped_fielded_fpath,
            },
            out_paths={
                'aligned_imgonly_fpath': aligned_imgonly_fpath,
            },
            stage='align_kwcoco',
        )

        # TODO:
        # Project annotation from latest annotations subdir
        # Prepare splits
        # Add baseline datasets to DVC

        aligned_viz_dpath = aligned_kwcoco_bundle / '_viz512'
        viz_max_dim = 512

        if config['visualize']:
            pipeline.submit(command=ub.codeblock(
                rf'''
                python -m watch visualize \
                    --src "{aligned_imgonly_fpath}" \
                    --viz_dpath "{aligned_viz_dpath}" \
                    --draw_anns=False \
                    --draw_imgs=True \
                    --channels="red|green|blue" \
                    --max_dim={viz_max_dim} \
                    --stack=only \
                    --animate=True --workers=auto
                '''), depends=[align_job], name=f'viz-imgs-{name}',
                in_paths={
                    'aligned_imgonly_fpath': aligned_imgonly_fpath,
                },
                out_paths={
                    'aligned_viz_dpath': aligned_viz_dpath,
                },
                stage='viz_imgs',
            )

        if site_globstr:
            # site_model_dpath = (dvc_dpath / 'annotations/site_models').shrinkuser(home='$HOME')
            # region_model_dpath = (dvc_dpath / 'annotations/region_models').shrinkuser(home='$HOME')

            # Visualization here is too slow, add on another option if we really
            # need to
            # viz_part = '--viz_dpath=auto' if config.visualize else ''
            viz_part = ''
            project_anns_job = pipeline.submit(command=ub.codeblock(
                rf'''
                python -m watch reproject_annotations \
                    --src "{aligned_imgonly_fpath}" \
                    --dst "{aligned_imganns_fpath}" \
                    --propogate_strategy="{config.propogate_strategy}" \
                    --site_models="{site_globstr}" \
                    --region_models="{region_globstr}" {viz_part}
                '''), depends=[align_job], name=f'project-annots-{name}',
                in_paths={
                    'aligned_imgonly_fpath': aligned_imgonly_fpath,
                },
                out_paths={
                    'aligned_imganns_fpath': aligned_imganns_fpath,
                },
                stage='project_annots',
            )

            if config.visualize:
                pipeline.submit(command=ub.codeblock(
                    rf'''
                    python -m watch visualize \
                        --src "{aligned_imganns_fpath}" \
                        --viz_dpath "{aligned_viz_dpath}" \
                        --draw_anns=True \
                        --draw_imgs=False \
                        --channels="red|green|blue" \
                        --max_dim={viz_max_dim} \
                        --animate=True --workers=auto \
                        --stack=only \
                        --only_boxes={config["visualize_only_boxes"]}
                    '''), depends=[project_anns_job], name=f'viz-annots-{name}',
                    in_paths={
                        'aligned_imganns_fpath': aligned_imganns_fpath,
                    },
                    out_paths={
                        'aligned_viz_dpath': aligned_viz_dpath,
                    },
                    stage='viz_anns',
                )
        else:
            aligned_imganns_fpath = aligned_imgonly_fpath
            info['aligned_imganns_fpath'] = aligned_imgonly_fpath
            project_anns_job = align_job

        align_info = info.copy()
        align_info['job'] = project_anns_job
        alignment_jobs.append(align_info)

    if config.separate_align_jobs:
        # Need a final union step
        aligned_fpaths = [d['aligned_imganns_fpath'] for d in alignment_jobs]
        union_depends_jobs = [d['job'] for d in alignment_jobs]
        # union_suffix = ub.hash_data([p.name for p in aligned_fpaths])[0:8]
        aligned_final_fpath = (aligned_kwcoco_bundle / 'data.kwcoco.json').shrinkuser(home='$HOME')
        aligned_multi_src_part = ' '.join(['"{}"'.format(p) for p in aligned_fpaths])
        # cache_prefix = f'[[ -f {aligned_final_fpath} ]] || ' if stages.nodes['final_union']['cache'] else ''
        union_job = pipeline.submit(command=ub.codeblock(
            rf'''
            # COMBINE Uncropped datasets
            {job_environ_str}python -m kwcoco union \
                --src {aligned_multi_src_part} \
                --dst "{aligned_final_fpath}"
            '''), depends=union_depends_jobs, name='kwcoco-union',
            in_paths={
                'aligned_fpaths': aligned_fpaths,
            },
            out_paths={
                'aligned_final_fpath': aligned_final_fpath,
            },
            stage='final_union',
        )
        aligned_final_jobs = [union_job]
    else:
        assert len(alignment_jobs) == 1
        aligned_final_fpath = alignment_jobs[0]['aligned_imganns_fpath']
        aligned_final_fpath = alignment_jobs[0]['aligned_imganns_fpath']
        aligned_final_jobs = [alignment_jobs[0]['job']]

    # TODO:
    # queue.synchronize -
    # force all submissions to finish before starting new ones.

    # Determine what stages will be cached.
    cache = config.cache
    if isinstance(cache, str):
        cache = [p.strip() for p in cache.split(',')]

    self = pipeline
    pipeline._update_stage_otf_cache(cache)

    config.tmux_workers = min(len(stac_jobs), config.tmux_workers)
    queue = config.create_queue(environ=environ)
    # queue = cmd_queue.Queue.create(
    #     backend=config['backend'], name=config['queue_name'], size=1,
    #     gres=None, environ=environ)

    # self._populate_explicit_dependency_queue(queue)
    self._populate_implicit_dependency_queue(
        queue, skip_existing=config.skip_existing)

    # queue.print_graph()
    # queue.rprint()

    # Do Basic Splits
    if config.splits:
        # Note: we probably could just do unions more cleverly rather than
        # splitting.
        from watch.cli import prepare_splits
        prepare_splits._submit_split_jobs(
            aligned_final_fpath, queue, depends=[aligned_final_jobs])

    # TODO: Can start the DVC add of the region subdirectories here
    ub.codeblock(
        '''
        7z a splits.zip data*.kwcoco.json

        ls */WV
        ls */L8
        ls */S2
        ls */*.json
        dvc add *.zip
        dvc add */WV */L8 */S2 */*.json *.zip

        DVC_DPATH=$(geowatch_dvc)
        echo "DVC_DPATH='$DVC_DPATH'"

        cd $DVC_DPATH/
        git pull  # ensure you are up to date with master on DVC
        cd $DVC_DPATH/Aligned-Drop3-TA1-2022-03-10
        dvc pull */L8.dvc */S2.dvc
        dvc pull
        */*.json
        ''')

    # pipeline.submit(ub.codeblock(
    #     '''
    #     DVC_DPATH=$(geowatch_dvc)
    #     python -m watch.cli.prepare_splits \
    #         --base_fpath=$DVC_DPATH/Drop2-Aligned-TA1-2022-02-15/data.kwcoco.json \
    #         --run=1 --serial=True
    #     '''))

    config.run_queue(queue)

    # if config.rprint:
    #     queue.print_graph()
    #     queue.rprint()
    # if config.run:
    #     # This logic will exist in cmd-queue itself
    #     other_session_handler = config.other_session_handler
    #     queue.run(block=True, system=True, with_textual=config.with_textual,
    #               other_session_handler=other_session_handler)

    # TODO: team features
    """
    DATASET_CODE=Aligned-Drop2-TA1-2022-03-07
    DVC_DPATH=$(geowatch_dvc)
    DATASET_CODE=Drop2-Aligned-TA1-2022-02-15
    KWCOCO_BUNDLE_DPATH=$DVC_DPATH/$DATASET_CODE
    python -m watch.cli.prepare_teamfeats \
        --base_fpath=$KWCOCO_BUNDLE_DPATH/data.kwcoco.json \
        --gres=0, \
        --with_depth=0 \
        --with_landcover=0 \
        --with_invariants=0 \
        --with_materials=1 \
        --depth_workers=auto \
        --do_splits=1  --cache=1 --run=0
    """

# dvc_dpath=$home/data/dvc-repos/smart_watch_dvc
# #dvc_dpath=$(geowatch_dvc)
# #s3_dpath=s3://kitware-smart-watch-data/processed/ta1/eval2/master_collation_working
# query_basename=$(basename "$s3_fpath")
# aligned_bundle_name=aligned-$dataset_suffix
# uncropped_bundle_name=uncropped-$dataset_suffix

# #region_models=$dvc_dpath'/annotations/region_models/*.geojson'
# region_models=$dvc_dpath/annotations/region_models/kr_r002.geojson
# # helper variables
# uncropped_dpath=$dvc_dpath/$uncropped_bundle_name
# uncropped_query_dpath=$uncropped_dpath/_query/items
# uncropped_ingress_dpath=$uncropped_dpath/ingress
# uncropped_kwcoco_fpath=$uncropped_dpath/data.kwcoco.json
# aligned_kwcoco_bundle=$dvc_dpath/$aligned_bundle_name
# aligned_kwcoco_fpath=$aligned_kwcoco_bundle/data.kwcoco.json
# uncropped_query_fpath=$uncropped_query_dpath/$query_basename
# uncropped_catalog_fpath=$uncropped_ingress_dpath/catalog.json

# export AWS_DEFAULT_PROFILE=iarpa
#     pass


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/watch/cli/prepare_ta2_dataset.py
    """
    main(cmdline=True)



import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from argparse import Namespace
import pytorch_lightning as pl
from torchmetrics.classification.accuracy import Accuracy
from tqdm import tqdm
from torch import pca_lowrank as pca

import json
import torch.package
import ubelt as ub
import os

from .data.datasets import kwcoco_dataset, gridded_dataset
from .utils.attention_unet import attention_unet


class pretext(pl.LightningModule):
    TASK_NAMES = [
        'sort',
        'augment',
        'overlap'
    ]

    def __init__(self, hparams):
        super().__init__()

        if type(hparams) == dict:
            hparams = Namespace(**hparams)

        self.save_hyperparameters(hparams)

        if hparams.train_dataset is not None:
            # self.trainset = gridded_dataset(hparams.train_dataset, sensor=hparams.sensor, bands=hparams.bands, patch_size=hparams.patch_size)
            self.trainset = kwcoco_dataset(hparams.train_dataset, sensor=hparams.sensor, bands=hparams.bands, patch_size=hparams.patch_size)
        else:
            self.trainset = None
        if hparams.vali_dataset is not None:
            self.valset = gridded_dataset(hparams.vali_dataset, sensor=hparams.sensor, bands=hparams.bands, patch_size=hparams.patch_size)
        else:
            self.valset = None

        print('hparams = {!r}'.format(hparams))
        if hasattr(hparams, 'num_channels'):
            # hack for loading without dataset state
            num_channels = hparams.num_channels
        elif self.trainset is not None:
            num_channels = self.trainset.num_channels
        else:
            num_channels = 6

        # determine which tasks to run
        self.task_indices = []
        # no tasks specified
        if len(self.hparams.tasks) < 1:
            raise ValueError(f'tasks must be specified. Options are {", ".join(pretext.TASK_NAMES)}, or all')
        # perform all tasks
        elif len(self.hparams.tasks) == 1 and self.hparams.tasks[0].lower() == 'all':
            self.task_indices = [i for i in range(len(pretext.TASK_NAMES))]
        # run a subset of tasks
        else:
            for task in self.hparams.tasks:
                if task.lower() in pretext.TASK_NAMES:
                    self.task_indices.append(pretext.TASK_NAMES.index(task.lower()))
                else:
                    raise ValueError(f'\'{task}\' not recognized as an available task. Options are {", ".join(pretext.TASK_NAMES)}, or all')
        if pretext.TASK_NAMES.index('augment') in self.task_indices and pretext.TASK_NAMES.index('overlap') not in self.task_indices:
            self.task_indices.append(pretext.TASK_NAMES.index('overlap'))

        # shared model body
        self.backbone = attention_unet(in_channels=num_channels,
                                        out_channels=hparams.feature_dim_shared,
                                        pos_encode=hparams.positional_encoding,
                                        attention_layers=hparams.attention_layers,
                                        mode=hparams.positional_encoding_mode)

        # task specific necks
        self.necks = [
            self.task_neck(2 * self.hparams.feature_dim_shared, self.hparams.feature_dim_each_task),  # sort task
            self.task_neck(self.hparams.feature_dim_shared, self.hparams.feature_dim_each_task),  # augment task
            self.task_neck(self.hparams.feature_dim_shared, self.hparams.feature_dim_each_task),  # overlap task
        ]
        self.necks = nn.ModuleList([ self.necks[i] for i in self.task_indices])

        # task specific heads
        self.heads = [
            self.pixel_classification_head(self.hparams.feature_dim_each_task, num_classes=2),  # sort task
            self.image_classification_head( self.hparams.feature_dim_each_task),  # augment task
            self.image_classification_head( self.hparams.feature_dim_each_task),  # overlap task
        ]
        self.heads = nn.ModuleList([ self.heads[i] for i in self.task_indices ])
        # task specific criterion
        self.criteria = [
            # BinaryFocalLoss(gamma=self.hparams.focal_gamma),  # sort task
            nn.NLLLoss(),  # sort task
            nn.TripletMarginLoss(),  # augment task
            nn.TripletMarginLoss(),  # overlap task
        ]
        self.criteria = [ self.criteria[i] for i in self.task_indices ]

        # task specific metrics
        self.sort_accuracy = Accuracy(task='binary')

    def forward(self, image_stack, positional_encoding=None):
        # pass through shared model body
        return torch.tanh(self.backbone(image_stack, positional_encoding))

    def shared_step(self, batch):
        # get features of each image from shared model body
        image_stack = torch.stack([batch['image1'], batch['image2'], batch['offset_image1'], batch['augmented_image1']], dim=1).to(self.device).float()
        # positional_encoding must be set to none to produce viable pretext task results
        out = self(image_stack, positional_encoding=None)
        image1_features = out[:, 0, :, :, :]
        image2_features = out[:, 1, :, :, :]
        offset_image1_features = out[:, 2, :, :, :]
        augmented_image1_features = out[:, 3, :, :, :]
        # get time sort labels
        time_labels = batch['time_sort_label']

        time_sort_labels = time_labels.unsqueeze(1).unsqueeze(1).repeat(1, image_stack.shape[-2], image_stack.shape[-1]).to(self.device)
        # time_sort_labels = 99 * torch.ones_like(time_labels)
        # time_sort_labels[:, self.hparams.ignore_boundary:-self.hparams.ignore_boundary, self.hparams.ignore_boundary:-self.hparams.ignore_boundary] = time_labels[:, self.hparams.ignore_boundary:-self.hparams.ignore_boundary, self.hparams.ignore_boundary:-self.hparams.ignore_boundary]

        losses = []
        output = {}

        # Time Sort task
        if 0 in self.task_indices:
            module_list_idx = self.task_indices.index(0)
            time_sort_prediction = self.necks[module_list_idx](torch.cat((image1_features, image2_features), dim=1))
            time_sort_prediction = self.heads[module_list_idx](time_sort_prediction)
            # evaluate
            loss_time = self.criteria[module_list_idx](time_sort_prediction.flatten(2).float(), time_sort_labels.flatten(1).long())
            time_accuracy = self.sort_accuracy((torch.max(time_sort_prediction.data, 1)[1]), time_sort_labels.int())

            losses.append(loss_time)
            output['time_accuracy'] = time_accuracy
            output['loss_time_sort'] = loss_time.detach()
            output['before_after_heatmap'] = time_sort_prediction
        else:
            output['before_after_heatmap'] = None

        # Overlap task
        if 2 in self.task_indices:
            module_list_idx = self.task_indices.index(2)
            # forward pass through neck
            image1_overlap_out = self.necks[module_list_idx](image1_features)
            image2_overlap_out = self.necks[module_list_idx](image2_features)
            image1_offset_overlap_out = self.necks[module_list_idx](offset_image1_features)
            # forward pass through head
            image1_overlap_out = self.heads[module_list_idx](image1_overlap_out)
            image2_overlap_out = self.heads[module_list_idx](image2_overlap_out)
            image1_offset_overlap_out = self.heads[module_list_idx](image1_offset_overlap_out)
            # evaluate
            loss_offset = self.criteria[module_list_idx](image1_overlap_out, image2_overlap_out, image1_offset_overlap_out)

            losses.append(loss_offset)
            if 2 in self.task_indices:
                output['loss_offset'] = loss_offset.detach()

        # Augment task
        if 1 in self.task_indices:
            module_list_idx = self.task_indices.index(1)
            # image1 forward pass through neck
            image1_augment_out = self.necks[module_list_idx](image1_features)
            image1_augmented_augment_out = self.necks[module_list_idx](augmented_image1_features)
            # image1 forward pass through head
            image1_augment_out = self.heads[module_list_idx](image1_augment_out)
            image1_augmented_augment_out = self.heads[module_list_idx](image1_augmented_augment_out)
            # perform computation if not already done
            if 2 not in self.task_indices:
                image1_offset_overlap_out = self.heads[module_list_idx](image1_offset_overlap_out)
            # image1 evaluate
            loss_augment = self.criteria[module_list_idx](image1_augment_out, image1_augmented_augment_out, image1_offset_overlap_out)

            losses.append(loss_augment)
            output['loss_augmented'] = loss_augment.detach()

        # add up loss
        loss = torch.sum(torch.stack(losses))
        output['loss'] = loss

        return output

    def training_step(self, batch, batch_idx):
        output = self.shared_step(batch)
        output.pop('before_after_heatmap')
        self.log_dict(output)
        return output

    def validation_step(self, batch, batch_idx):
        output = self.shared_step(batch)
        output.pop('before_after_heatmap')
        output = {"val_" + key: val for key, val in output.items()}
        self.log_dict(output)
        return output

    def predict(self, batch):
        self.eval()
        feats = {}
        shared_feats = self(batch)
        feats['shared'] = shared_feats
        return feats['shared']

    def train_dataloader(self):
        return DataLoader(self.trainset, batch_size=self.hparams.batch_size, num_workers=self.hparams.workers, shuffle=True, pin_memory=False)

    def val_dataloader(self):
        if self.hparams.vali_dataset is not None:
            return DataLoader(self.valset, batch_size=self.hparams.batch_size, num_workers=self.hparams.workers, pin_memory=False)

    def configure_optimizers(self):
        opt = torch.optim.AdamW(
            self.parameters(),
            lr=self.hparams.learning_rate)

        lr_scheduler = torch.optim.lr_scheduler.StepLR(
            opt, step_size=self.hparams.step_size,
            gamma=self.hparams.lr_gamma)

        return {'optimizer': opt, 'lr_scheduler': lr_scheduler}

    def task_neck(self, in_chan, out_chan):
        kernel_size = 3
        stride = 1
        padding = int((kernel_size - 1) / 2)
        return nn.Sequential(
            nn.BatchNorm2d(in_chan),
            nn.ReLU(inplace=True),
            nn.Conv2d(in_chan, out_chan, kernel_size, stride, padding),
            nn.BatchNorm2d(out_chan),
            nn.ReLU(inplace=True)
        )

    def pixel_classification_head(self, in_chan, num_classes):
        kernel_size = 1
        stride = 1
        padding = int((kernel_size - 1) / 2)
        return nn.Sequential(
                            nn.Conv2d(in_chan, in_chan, kernel_size, stride, padding),
                            nn.BatchNorm2d(in_chan),
                            nn.ReLU(inplace=True),
                            nn.Conv2d(in_chan, in_chan, kernel_size, stride, padding),
                            nn.BatchNorm2d(in_chan),
                            nn.ReLU(inplace=True),
                            nn.Conv2d(in_chan, num_classes, 1, bias=False, padding=0),
                            nn.LogSoftmax(dim=1)
                            )

    def image_classification_head(self, in_chan):
        return nn.Sequential(
                             nn.Flatten(1, -1)
                            )

    # def image_classification_head(self, in_chan):
    #     return nn.Sequential(
    #                          nn.AdaptiveAvgPool2d(output_size=(1, 1)),
    #                          nn.Flatten(1, -1),
    #                          nn.Linear(in_chan, 8),
    #                         )

    def generate_pca_matrix(self, save_path, loader, reduction_dim=6):
        feature_collection = []

        with torch.set_grad_enabled(False):
            # TODO: option to cache or specify a specific projection matrix?
            for n, batch in tqdm(enumerate(loader), desc='Calculating PCA matrix', total=50):
                if n == 50:
                    break
                image_stack = torch.stack([batch['image1'], batch['image2'], batch['offset_image1'], batch['augmented_image1']], dim=1)
                features = self.forward(image_stack.to(self.device))
                feature_collection.append(features.cpu())

            features = None
            image_stack = None
            stack = torch.cat(feature_collection, dim=0).permute(0, 1, 3, 4, 2).reshape(-1, self.hparams.feature_dim_shared)
            _, _, projector = pca(stack, q=reduction_dim)
            stack = None

        projector = projector.permute(1, 0)

        if save_path:
            torch.save(projector, save_path)

        return projector

    def on_save_checkpoint(self, checkpoint):
        if self.hparams.pca_projection_path:
            save_dir = self.hparams.pca_projection_path
            save_path = os.path.join(save_dir, 'pretext_pca_{}'.format(str(self.current_epoch)) + '.pt')
            os.makedirs(save_dir, exist_ok=True)
            self.generate_pca_matrix(save_path=save_path, loader=self.train_dataloader(), reduction_dim=self.hparams.reduction_dim)

    def save_package(self, package_path):
        model = self

        package_path = os.path.join(package_path, 'pretext_package.pt')

        backup_attributes = {}
        unsaved_attributes = [
            'trainer',
            'train_dataloader',
            'val_dataloader',
            'test_dataloader',
            '_load_state_dict_pre_hooks',
        ]
        for key in unsaved_attributes:

            val = getattr(model, key)
            #print(val)
            if val is not None:
                backup_attributes[key] = val

        log_path = package_path
        log_path = ub.Path(log_path)

        metadata_fpaths = []

        metadata_fpaths += list(log_path.glob('hparams.yaml'))
        try:
            for key in backup_attributes.keys():
                setattr(model, key, None)
            arch_name = 'pretext_model.pkl'
            module_name = 'watch_tasks_invariants'

            with torch.package.PackageExporter(package_path) as exp:
                exp.extern('**', exclude=['watch.tasks.invariants.**'])
                exp.intern('watch.tasks.invariants.**', allow_empty=False)

                package_header = {
                    'version': '0.1.0',
                    'arch_name': arch_name,
                    'module_name': module_name,
                }
                exp.save_text(
                    'package_header', 'package_header.json',
                    json.dumps(package_header)
                )
                exp.save_pickle(module_name, arch_name, model)
                for meta_fpath in metadata_fpaths:
                    with open(meta_fpath, 'r') as file:
                        text = file.read()
                    exp.save_text('package_header', meta_fpath.name, text)
        finally:

            for key, val in backup_attributes.items():
                setattr(model, key, val)

    @classmethod
    def load_package(cls, package_path):
        """
        DEPRECATE IN FAVOR OF watch.tasks.fusion.utils.load_model_from_package

        TODO:
            - [ ] Make the logic that defines the save_package and load_package
                methods with appropriate package header data a lightning
                abstraction.
        """
        # NOTE: there is no gaurentee that this loads an instance of THIS
        # model, the model is defined by the package and the tool that loads it
        # is agnostic to the model contained in said package.
        # This classmethod existing is a convinience more than anything else
        from watch.tasks.fusion.utils import load_model_from_package

        self = load_model_from_package(package_path)
        return self



#!/usr/bin/env python3
"""
See Old Version:
    ../../../scripts/run_sc_fusion_eval3_for_baseline.py

SeeAlso:
    ~/code/watch-smartflow-dags/KIT_TA2_PREEVAL10_PYENV.py
"""
import json
import os
import shutil
import subprocess

import scriptconfig as scfg
import ubelt as ub
from watch.mlops.smart_pipeline import DinoBoxDetector, DinoBuildingFilter

from glob import glob


class DinoSVConfig(scfg.DataConfig):
    """
    Run TA-2 SC fusion as baseline framework component
    """
    input_path = scfg.Value(None, type=str, position=1, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework JSON
            '''))
    input_region_path = scfg.Value(None, type=str, position=2, required=True, help=ub.paragraph(
            '''
            Path to input T&E Baseline Framework Region definition JSON
            '''))
    output_path = scfg.Value(None, type=str, position=3, required=True, help='S3 path for output JSON')
    aws_profile = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            AWS Profile to use for AWS S3 CLI commands
            '''))
    dryrun = scfg.Value(False, isflag=True, short_alias=['d'], help='Run AWS CLI commands with --dryrun flag')
    outbucket = scfg.Value(None, type=str, required=True, short_alias=['o'], help=ub.paragraph(
            '''
            S3 Output directory for STAC item / asset egress
            '''))
    ta2_s3_collation_bucket = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            S3 Location for collated TA-2 output (bucket name should
            include up to eval name)
            '''))
    dino_detect_config = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            Raw json/yaml or a path to a json/yaml file that specifies the
            config for DinoBoxDetector.
            '''))
    dino_filter_config = scfg.Value(None, type=str, help=ub.paragraph(
            '''
            Raw json/yaml or a path to a json/yaml file that specifies the
            config for DinoBuildingFilter.
            '''))


def main():
    config = DinoSVConfig.cli(strict=True)
    print('config = {}'.format(ub.urepr(dict(config), nl=1, align=':')))
    run_dino_sv(config)


def _upload_region(aws_base_command,
                   local_region_dir,
                   local_input_region_path,
                   destination_region_s3):
    with open(local_input_region_path) as f:
        region = json.load(f)

    region_id = None
    for feature in region.get('features', ()):
        props = feature['properties']
        if props['type'] == 'region':
            region_id = props.get('region_model_id', props.get('region_id'))
            break

    if region_id is not None:
        updated_region_path = os.path.join(local_region_dir,
                                           '{}.geojson'.format(region_id))

        print("** Uploading updated region file")
        subprocess.run([*aws_base_command,
                        updated_region_path, destination_region_s3],
                       check=True)
    else:
        print("** Error: Couldn't parse region_id from region file, "
              "not uploading")


def _ta2_collate_output(aws_base_command,
                        local_region_dir,
                        local_sites_dir,
                        destination_s3_bucket,
                        performer_suffix='KIT'):
    def _get_suffixed_basename(local_path):
        base, ext = os.path.splitext(os.path.basename(local_path))
        return "{}_{}{}".format(base, performer_suffix, ext)

    for region in glob(os.path.join(local_region_dir, '*.geojson')):
        region_s3_outpath = '/'.join((destination_s3_bucket,
                                      'region_models',
                                      _get_suffixed_basename(region)))
        subprocess.run([*aws_base_command,
                        region,
                        region_s3_outpath], check=True)

    for site in glob(os.path.join(local_sites_dir, '*.geojson')):
        site_s3_outpath = '/'.join((destination_s3_bucket,
                                    'site_models',
                                    _get_suffixed_basename(site)))
        subprocess.run([*aws_base_command,
                        site,
                        site_s3_outpath], check=True)


def run_dino_sv(config):
    from watch.cli.baseline_framework_kwcoco_ingress import baseline_framework_kwcoco_ingress
    from watch.cli.baseline_framework_kwcoco_egress import baseline_framework_kwcoco_egress
    from watch.utils.util_framework import download_region, determine_region_id
    from watch.utils.util_yaml import Yaml

    input_path = config.input_path
    input_region_path = config.input_region_path
    output_path = config.output_path

    outbucket = config.outbucket
    ta2_s3_collation_bucket = config.ta2_s3_collation_bucket
    aws_profile = config.aws_profile
    dryrun = config.dryrun

    dino_detect_config = config.dino_detect_config
    dino_filter_config = config.dino_filter_config

    if aws_profile is not None:
        aws_base_command =\
            ['aws', 's3', '--profile', aws_profile, 'cp']
    else:
        aws_base_command = ['aws', 's3', 'cp']

    if dryrun:
        aws_base_command.append('--dryrun')

    # 1. Ingress data
    print("* Running baseline framework kwcoco ingress *")
    ingress_dir = '/tmp/ingress'
    ingress_kwcoco_path = baseline_framework_kwcoco_ingress(
        input_path,
        ingress_dir,
        aws_profile,
        dryrun)

    # # 2. Download and prune region file
    print("* Downloading and pruning region file *")
    local_region_path = '/tmp/region.json'

    download_region(
        input_region_path=input_region_path,
        output_region_path=local_region_path,
        aws_profile=aws_profile,
        strip_nonregions=True,
    )

    # Determine the region_id in the region file.
    region_id = determine_region_id(local_region_path)

    dino_boxes_kwcoco_path = os.path.join(
        ingress_dir, 'dino_boxes_kwcoco.json')

    cropped_region_models_bas = os.path.join(ingress_dir,
                                             'cropped_region_models_bas')
    cropped_site_models_bas = os.path.join(ingress_dir,
                                           'cropped_site_models_bas')

    site_models_outdir = os.path.join(ingress_dir, 'sv_out_site_models')
    os.makedirs(site_models_outdir, exist_ok=True)
    region_models_outdir = os.path.join(ingress_dir, 'sv_out_region_models')
    os.makedirs(region_models_outdir, exist_ok=True)

    site_models_manifest_outdir = os.path.join(
        ingress_dir, 'tracking_manifests_sv')
    os.makedirs(site_models_manifest_outdir, exist_ok=True)
    site_models_manifest_outpath = os.path.join(
        site_models_manifest_outdir, 'site_models_manifest.json')
    # Copy input region model into region_models outdir to be updated
    # (rather than generated from tracking, which may not have the
    # same bounds as the original)
    shutil.copy(local_region_path, os.path.join(
        region_models_outdir, '{}.geojson'.format(region_id)))

    # 3.1. Check that we have at least one "video" (BAS identified
    # site) to run over; if not skip SV fusion and KWCOCO to GeoJSON
    with open(ingress_kwcoco_path) as f:
        ingress_kwcoco_data = json.load(f)

    if len(ingress_kwcoco_data.get('videos', ())) > 0:
        # 3.2 Run DinoBoxDetector
        print("* Running Dino Detect *")

        default_dino_detect_config = ub.udict({
            'coco_fpath': ingress_kwcoco_path,
            'package_fpath': None,
            'batch_size': 1,
            'device': 0})
        dino_detect_config = (default_dino_detect_config
                              | Yaml.coerce(dino_detect_config or {}))

        if dino_detect_config.get('package_fpath', None) is None:
            raise ValueError('Requires package_fpath')

        dino_box_detector = DinoBoxDetector(root_dpath='/tmp/ingress')
        dino_box_detector.configure({
            'out_coco_fpath': dino_boxes_kwcoco_path,
            **dino_detect_config})

        ub.cmd(dino_box_detector.command(), check=True, verbose=3, system=True)

        # 3.3 Run DinoBuildingFilter
        print("* Running Dino Building Filter *")

        default_dino_filter_config = ub.udict({})
        dino_filter_config = (default_dino_filter_config
                              | Yaml.coerce(dino_filter_config or {}))

        input_region_model_bas = os.path.join(cropped_region_models_bas,
                                              '{}.geojson'.format(region_id))
        output_region_model = os.path.join(region_models_outdir,
                                           '{}.geojson'.format(region_id))
        dino_building_filter = DinoBuildingFilter(root_dpath='/tmp/ingress')
        dino_building_filter.configure({
            'input_kwcoco': dino_boxes_kwcoco_path,
            'input_region': input_region_model_bas,
            'input_sites': cropped_site_models_bas,
            'output_region_fpath': output_region_model,
            'output_sites_dpath': site_models_outdir,
            'output_site_manifest_fpath': site_models_manifest_outpath,
            })

        ub.cmd(dino_building_filter.command(), check=True, verbose=3, system=True)

    # 4. (Optional) collate TA-2 output
    if ta2_s3_collation_bucket is not None:
        print("* Collating TA-2 output")
        _ta2_collate_output(aws_base_command,
                            region_models_outdir,
                            site_models_outdir,
                            ta2_s3_collation_bucket)

    # 5. Egress (envelop KWCOCO dataset in a STAC item and egress;
    #    will need to recursive copy the kwcoco output directory up to
    #    S3 bucket)
    print("* Egressing KWCOCO dataset and associated STAC item *")
    baseline_framework_kwcoco_egress(dino_boxes_kwcoco_path,
                                     local_region_path,
                                     output_path,
                                     outbucket,
                                     aws_profile=None,
                                     dryrun=False,
                                     newline=False)


if __name__ == "__main__":
    main()



r"""
Split a coco dataset into one per video.

Ignore:
    DVC_DATA_DPATH=$(geowatch_dvc --tags='phase2_data' --hardware=auto)
    python -m watch.cli.split_videos \
        --src "$DVC_DATA_DPATH/Drop4-BAS/data_train.kwcoco.json" \
              "$DVC_DATA_DPATH/Drop4-BAS/data_vali.kwcoco.json" \
        --dst_dpath "$DVC_DATA_DPATH/Drop4-BAS/"

    python -m watch.cli.split_videos \
        --dst_dpath "$DVC_DATA_DPATH/Drop4-BAS/"

"""
import scriptconfig as scfg


class SplitVideoConfig(scfg.DataConfig):
    """
    Breaks one or more kwcoco file containing multiple videos into single
    kwcoco files per video. The new kwcoco file names use the same name as the
    input dataset, but prefix it with the video name.
    """
    src = scfg.Value(None, nargs='+', help='one or more datasets to split', position=1)

    dst_dpath = scfg.Value(None, help=(
        'path to write to. If None, uses the src dataset path'))

    dst = scfg.Value(None, help=(
        'If specified, this is a format string template for the path that '
        'will be written to. An error will be thrown if they are not unique.'
        'Available keys are {src_name} and {video_name} in this version.'
        'This name will be relative to dst_dpath. Defaults to '
        '``{src_name}_{video_name}.kwcoco.json``'
    ))

    io_workers = scfg.Value(0, help='number of background IO workers')


def main(cmdline=1, **kwargs):
    """
    Ignore:
        from watch.cli.split_videos import *  # NOQA
        import watch
        data_dvc_dpath = watch.find_dvc_dpath(tags='phase2_data', hardware='auto')
        src_fpath = [
            str(data_dvc_dpath / 'Drop4-BAS/data_train.kwcoco.json'),
            str(data_dvc_dpath / 'Drop4-BAS/data_vali.kwcoco.json'),
        ]
        from watch.utils import util_pattern
        coco_fpaths = list(util_pattern.MultiPattern.coerce(list(map(str, src_fpath))) .paths())
        print(f'coco_fpaths={coco_fpaths}')
        cmdline = 0
        kwargs = dict(src=src_fpath)
    """
    import kwcoco
    import ubelt as ub
    from watch.utils import util_pattern
    from watch.utils import util_parallel
    config = SplitVideoConfig.cli(cmdline=cmdline, data=kwargs)
    print('config = {}'.format(ub.urepr(config, nl=1)))

    coco_fpaths = list(util_pattern.MultiPattern.coerce(config.src).paths())
    print(f'coco_fpaths={coco_fpaths}')

    io_workers = util_parallel.coerce_num_workers(config.io_workers)
    writer = util_parallel.BlockingJobQueue(max_workers=io_workers)

    writen_fpaths = set()

    if config['dst'] is None:
        dst_template = '{src_name}_{video_name}.kwcoco.json'
    else:
        dst_template = config['dst']
        assert '{' in dst_template
        assert '}' in dst_template

    for coco_fpath in coco_fpaths:
        print(f'splitting coco_fpath={coco_fpath}')
        dset = kwcoco.CocoDataset.coerce(coco_fpath)

        src_name = ub.Path(dset.fpath).name.split('.')[0]

        for video in ub.ProgIter(dset.videos().objs, desc='Splitting dataset'):
            video_name = video['name']

            fmtkw = {
                'video_name': video_name,
                'src_name': src_name,
            }
            if config.dst_dpath is None:
                dst_dpath = dset.bundle_dpath
            else:
                dst_dpath = ub.Path(config.dst_dpath)
            vid_fpath = ub.Path(dst_dpath) / dst_template.format(**fmtkw)
            if vid_fpath in writen_fpaths:
                raise Exception('Split name template did not generate unique names')
            writen_fpaths.add(vid_fpath)
            # print(f'vidname={vidname}')
            video_gids = list(dset.images(video_id=video['id']))
            # print(f'video_gids={video_gids}')
            vid_subset = dset.subset(video_gids)
            vid_subset.fpath = vid_fpath
            writer.submit(vid_subset.dump)
            # vid_subset.dump(vid_subset.fpath, newlines=False)

        writer.wait_until_finished(desc="Finish write jobs")
        print(f'finished splitting coco_fpath={coco_fpath}')
    print('Finished splits')


if __name__ == '__main__':
    """
    CommandLine:
        python ~/code/watch/watch/cli/split_videos.py
    """
    main()
